"""
Simulate molecular reactions
"""

import itertools
from collections import namedtuple

import networkx as nx

from jdna.sequence import Sequence, SequenceFlags
from jdna.viewer import SequenceViewer
import primer3
from decimal import Decimal


class ReactionException(Exception):
    """Generic reaction exception"""


class PCRException(ReactionException):
    """Exception with pcr"""


BindingEvent = namedtuple("BindingEvent", ["template", "primer", "position"])


class Assembly(object):
    """Stores a DNA assembly.

    Assembly can be printed to display relevant assembly information:

    .. code::

        assembly.print()

    ::

        > "Unnamed" (320bp)
          Cyclic: True
          Num Fragments: 4
          Overhang Tms (°C): 53.65, 53.79, 54.96, 56.37
          Junction Lengths (bp): 20, 20, 20, 20
          Junction ΔG: -1.91e+4, -1.93e+4, -1.98e+4, -2.05e+4
          Junction ΔG (hairpin): (0.0, 232.50050218543765), (203.8205043739763, -84.3344956260189), (0.0, -60.91449781456504), (-1026.7134934374844, -622.0224934343823)
          Competing ΔG: -2.18e+4, -2.11e+4, -1.92e+4, -2.01e+4
          Product length (bp): 300

                  -----Tm: 53.7°C-----
        (0) 0     TTGACGACAGTGGCTATCCCCTGTTGCTAGGCACGGTGATATATCAGCCC
        (1) 0     --------------------------------------------------
        (2) 0     --------------------------------------------------
        (3) 0     --------------------------------------------------

                                           -----Tm: 53.8°C-----
        (0) 50    ATAGGGCCGGATCAACGTAGTTTGATCGTAGCTGTTCCGTCAGTC-----
        (1) 50    -------------------------TCGTAGCTGTTCCGTCAGTCAGTGC
        (2) 50    --------------------------------------------------
        (3) 50    --------------------------------------------------

        (0) 100   --------------------------------------------------
        (1) 100   CGATGCCAGTCTACTGCTTTTCGCCCAGGGGGACACCTGACACTATGTTA
        (2) 100   --------------------------------------------------
        (3) 100   --------------------------------------------------

                  -----Tm: 55.0°C-----
        (0) 150   --------------------------------------------------
        (1) 150   TGCCAAGCTCACGTTTCACA------------------------------
        (2) 150   TGCCAAGCTCACGTTTCACAAGGCAGATTAAATAAGACTGAAACATTTGG
        (3) 150   --------------------------------------------------

                                           -----Tm: 56.4°C-----
        (0) 200   --------------------------------------------------
        (1) 200   --------------------------------------------------
        (2) 200   TGGAGGCGCGAGTCTTGCCCTTAGGCAGACCTGTTCGAGGGCGAA-----
        (3) 200   -------------------------CAGACCTGTTCGAGGGCGAAAAACG

        (0) 250   --------------------------------------------------
        (1) 250   --------------------------------------------------
        (2) 250   --------------------------------------------------
        (3) 250   CCGAAGTCACATAGCGGCATGTTAGGTAGCTGTACACCCGTACTGCTACA
        """

    def __init__(self, templates, junctions, cyclic):
        """
        Assembly constructor

        :param templates: list of templates
        :type templates: list
        :param junctions: list of junctions
        :type junctions: list
        :param cyclic: whether assembly is cyclic
        :type cyclic: bool
        """
        self.templates = [t.copy() for t in templates]
        self.junctions = [o.copy() for o in junctions]
        self.cyclic = cyclic

    def tms(self):
        """Return the tms of the junctions"""
        return [round(o.tm(), 2) for o in self.junctions]

    @property
    def product(self):
        product = Sequence()
        product.name = "Unnamed assembly"
        if not self.cyclic:
            zipped = zip(self.templates, self.junctions)
        else:
            zipped = zip(self.junctions, self.templates)
        for s1, s2 in zipped:
            product += s1 + s2
        if self.cyclic:
            product.circularize()
        return product

    def view(self, *args, **kwargs):
        seqs = []
        positions = [0]
        pos = 0
        junctions = self.junctions[:]
        for jxn in junctions:
            jxn.annotate(None, None, "Tm: {}°C".format(round(jxn.tm(), 1)))
        if self.cyclic:
            junctions.append(junctions[0])
        else:
            junctions = [Sequence()] + junctions
        for i, t in enumerate(self.templates):
            seq = junctions[i] + t + junctions[i + 1]
            seqs.append(seq)
            pos += len(seq) - len(junctions[i + 1])
            positions.append(pos)

        aligned_seqs = []
        for p, s in zip(positions, seqs):
            aligned_seqs.append(Sequence("-" * p) + s)

        mx = max([len(s) for s in aligned_seqs])
        for i, s in enumerate(aligned_seqs):
            diff = mx - len(s)
            aligned_seqs[i] = aligned_seqs[i] + Sequence("-" * diff)

        labels = [
            "({i}) {index}".format(i=i, index="{index}")
            for i in range(len(aligned_seqs))
        ]
        viewer = SequenceViewer(
            aligned_seqs,
            *args,
            sequence_labels=labels,
            foreground_colors="RANDOM",
            **kwargs
        )
        for seq in aligned_seqs:
            Sequence._apply_features_to_view(seq, viewer)
        viewer.metadata["Cyclic"] = self.cyclic
        viewer.metadata["Num Fragments"] = len(self.templates)
        viewer.metadata["Fragment Names"] = "\n\t" + "\n\t".join(
            ["{} - {}".format(i, t.name) for i, t in enumerate(self.templates)]
        )
        viewer.metadata["Overhang Tms (°C)"] = ", ".join([str(x) for x in self.tms()])
        viewer.metadata["Junction Lengths (bp)"] = ", ".join(
            [str(len(x)) for x in self.junctions]
        )
        viewer.metadata["Junction ΔG"] = self.format_float_array(self.deltaGs())
        viewer.metadata["Junction ΔG (hairpin)"] = self.format_array(
            self.deltaG_hairpins()
        )
        viewer.metadata["Competing ΔG"] = self.format_float_array(
            self.competing_deltaGs()
        )
        viewer.metadata["Product length (bp)"] = "{}".format(len(self.product))
        return viewer

    @staticmethod
    def format_array(arr):
        return ", ".join([str(x) for x in arr])

    @staticmethod
    def format_float_array(arr):
        return ", ".join(["{:.2e}".format(Decimal(float(x))) for x in arr])

    def deltaG_hairpins(self):
        gs = []
        for jxn in self.junctions:
            fwd = self._hairpin(jxn).dg
            rev = self._hairpin(jxn.copy().rc()).dg
            gs.append((fwd, rev))
        return gs

    def _heterodimer(self, s1, s2):
        return primer3.calcHeterodimer(str(s1[:60]).upper(), str(s2[:60]).upper())

    def _hairpin(self, s):
        return primer3.calcHairpin(str(s[:60]).upper())

    def deltaGs(self):
        overhangs = [o[:60] for o in self.junctions]
        return [self._heterodimer(o, o.copy().rc()).dg for o in overhangs]

    def competing_deltaGs(self):
        gs = []
        for jxn in self.junctions:
            total = 0
            total += self._hairpin(jxn).dg
            total += self._hairpin(jxn.copy().rc()).dg
            others = self.junctions[:]
            others.remove(jxn)
            for other in others:
                dg1 = self._heterodimer(jxn, other.copy().rc()).dg
                dg2 = self._heterodimer(jxn, other.copy()).dg
                total += dg1
                total += dg2
            gs.append(total)
        return gs

    def print(self, *args, **kwargs):
        """
        Print the assembly.

        ::

            > "Unnamed" (320bp)
                  Num Fragments: 4
                  Overhang Tms (°C): 61.49, 58.37, 54.72, 45.91
                  Junction Lengths (bp): 20, 20, 20, 20
                  Junction ΔG: -2.30e+4, -2.18e+4, -1.97e+4, -1.54e+4
                  Junction ΔG (hairpin): (-1542.5069956260231, -1787.4319956260224), (-1832.7404978145642, -708.4919956260237), (0.0, 0.0), (0.0, 0.0)
                  Competing ΔG: -3.39e+4, -2.43e+4, -2.19e+4, -1.33e+4
                  Length: 300bp

                                                                                                     |<Tm: 58.4°C
                          -----Tm: 61.5°C-----                                                       ----------
                (0) 0     TGCTCGGCGAAGGGCCTGACAAGACCACTTCGTACCCTTTGTAACGTGACTTCTGTGAATCGATGGCGGATGTTGCTTGTGACGC
                (1) 0     ---------------------------------------------------------------------------CTTGTGACGC
                (2) 0     -------------------------------------------------------------------------------------
                (3) 0     -------------------------------------------------------------------------------------

                          |<Tm: 58.4°C
                          ----------                                                       -----Tm: 54.7°C-----
                (0) 85    ACCCGTAGCG---------------------------------------------------------------------------
                (1) 85    ACCCGTAGCGACATAACCGCCTTATTCCCACTTGTTTGGGGAGGCAAGTTTTGTAGTAGCCATTCTAATCCCGTTTTCTCCGCCG
                (2) 85    -----------------------------------------------------------------TAATCCCGTTTTCTCCGCCG
                (3) 85    -------------------------------------------------------------------------------------

                                                                                 -----Tm: 45.9°C-----
                (0) 170   -------------------------------------------------------------------------------------
                (1) 170   -------------------------------------------------------------------------------------
                (2) 170   CTTGTTCAGCTTGATGTGTTGGAACAGCAAGTTTATTTTAGGGTGTCAGGAGGGGGAGGTCCTCTAGTATTAAGC----------
                (3) 170   -------------------------------------------------------GAGGTCCTCTAGTATTAAGCACCGGCCAAG

                                                                       -----Tm: 61.5°C-----
                (0) 255   -----------------------------------------------------------------
                (1) 255   -----------------------------------------------------------------
                (2) 255   -----------------------------------------------------------------
                (3) 255   GCGGCACGGACGCCACTAGAAGGGAGGCGTTCATAGCGAGTCCTTTGCTCGGCGAAGGGCCTGAC

        """
        self.view(*args, **kwargs).print()

    def __str__(self):
        return str(self.view())


class Reaction(object):
    MIN_BASES = 13

    @classmethod
    def pcr(cls, template, primers, min_bases=MIN_BASES):
        """
        Make a pcr product from a template and primers.

        :param template: template
        :type template: Sequence
        :param primers: primers
        :type primers: list | tuple | iterable
        :param min_bases: minimum number of bases for primer annealing
        :type min_bases: int
        :return: list of Sequence instances
        :rtype: list
        """
        bindings = []
        for p in primers:
            bindings += template.anneal(p, min_bases=min_bases)

        forward_bindings = []
        reverse_bindings = []
        for bind in bindings:
            if bind.direction == SequenceFlags.FORWARD:
                forward_bindings.append(bind)
            elif bind.direction == SequenceFlags.REVERSE:
                reverse_bindings.append(bind)
            else:
                raise Exception("Direction {} not recognized".format(bind.direction))

        if not forward_bindings or not reverse_bindings:
            raise PCRException(
                "Some primers did not bind. Number of forward bindings: {}. Number of rev bindings: {}".format(
                    len(forward_bindings), len(reverse_bindings)
                )
            )
        products = []
        for fwd, rev in itertools.product(forward_bindings, reverse_bindings):
            overhang1 = fwd.five_prime_overhang
            overhang2 = rev.five_prime_overhang
            amplified = template.new_slice(fwd.start, rev.end)
            products.append(
                overhang1 + amplified + overhang2.copy().reverse_complement()
            )
        return products

    @classmethod
    def interaction_graph(
        cls,
        sequences,
        min_bases=Sequence.DEFAULTS.MIN_ANNEAL_BASES,
        bind_reverse_complement=False,
        max_bases=None,
        only_ends=False,
    ):
        """
        Make an interaction graph from a list of sequences

        :param sequences: list of sequences
        :type sequences: list
        :param min_bases: minimum number of bases to bind
        :type min_bases: int
        :param bind_reverse_complement: whether to reverse complement each sequence and check interactions
        :type bind_reverse_complement: int
        :param max_bases: max number of bases to search
        :type max_bases: int
        :param only_ends: whether to only consider end-to-end interactions
        :type only_ends: bool
        :return:
        :rtype:
        """
        G = nx.DiGraph()
        sequences = [s.copy() for s in sequences]
        if bind_reverse_complement:
            reverse_complement = []
            for s in sequences:
                new = s.copy().reverse_complement()
                new.name = s.name + "(-1)"
                reverse_complement.append(new)
            sequences += reverse_complement
        for s in sequences:
            G.add_node(s.global_id, sequence=s)
        seq_pairs = itertools.product(sequences, repeat=2)
        for s1, s2 in seq_pairs:
            for binding in s1.anneal_forward(s2, min_bases=min_bases, depth=max_bases):
                if only_ends:
                    if binding.span[0] != 0 or binding.query_span[-1] != len(s2) - 1:
                        break
                if not (binding.span[0] == 0 and binding.span[-1] == len(s1) - 1):
                    G.add_edge(
                        s2.global_id,
                        s1.global_id,
                        template=s1,
                        primer=s2,
                        binding=binding,
                    )
        return G

    @classmethod
    def interaction_report(cls, G):
        rows = []
        for n1, n2 in G.edges:
            edge = G.edges[n1, n2]
            binding = edge["binding"]
            template = edge["template"]
            primer = edge["primer"]
            template_info = "{name} (id={gid} ({start},{end}/{length})".format(
                name=template.name,
                gid=template.global_id,
                start=binding.span[0],
                end=binding.span[1],
                length=len(template) - 1,
            )
            primer_info = "{name} (id={gid} ({start},{end}/{length})".format(
                name=primer.name,
                gid=primer.global_id,
                start=binding.query_span[0],
                end=binding.query_span[1],
                length=len(primer) - 1,
            )
            rows.append("{} -> {}".format(template_info, primer_info))
        print("format: primer -> template")
        print("\n".join(rows))

    @classmethod
    def linear_paths(cls, G):
        """Find linear paths from an interaction graph"""
        paths = []
        subgraph = G.subgraph(G.nodes).copy()
        for _ in nx.simple_cycles(G):
            return []
        while subgraph.nodes:
            path = nx.dag_longest_path(subgraph)
            if len(path) < 2:
                break
            subgraph.remove_nodes_from(path)
            paths.append(path)
        return paths

    @classmethod
    def cyclic_paths(cls, G):
        """Find cyclic paths from an interaction graph"""
        paths = []
        for path in nx.simple_cycles(G.copy()):
            paths.append(path)
        return paths

    @classmethod
    def _path_to_edge_data(cls, G, path, cyclic):
        if cyclic:
            path_pairs = zip(path, path[1:] + [path[0]])
        else:
            path_pairs = zip(path[:-1], path[1:])
        edges = []
        for n1, n2 in path_pairs:
            edges.append(G.edges[n1, n2])
        return edges

    @classmethod
    def _path_to_assembly(self, path, G, cyclic=False):
        prev_template_end = None

        node_pairs = []
        overhangs = []

        for i, data in enumerate(self._path_to_edge_data(G, path, cyclic)):
            binding = data["binding"]

            # use the previous template and this query should refer to the same sequence
            node_pairs.append(
                (prev_template_end, binding.query_start.prev(), data["primer"].name)
            )

            # the next segment start is the template start
            prev_template_end = binding.end.next()
            overhangs.append(binding.template_anneal)

        if not cyclic:
            node_pairs.append((prev_template_end, None, data["template"].name))
            overhangs.append(Sequence())
        else:
            pass
            node_pairs[0] = (prev_template_end, node_pairs[0][1], node_pairs[0][2])
            overhangs = [overhangs[-1]] + overhangs[:-1]

        amplified_sequences = []
        for n1, n2, name in node_pairs:
            sequence = Sequence.new_slice(n1, n2)
            sequence.name = name
            amplified_sequences.append(sequence)

        return Assembly(amplified_sequences, overhangs, cyclic)

    @classmethod
    def linear_assemblies(
        cls, sequences, min_bases=Sequence.DEFAULTS.MIN_ANNEAL_BASES, max_bases=None
    ):
        """
        Finds all unique, longest linear assemblies. If cyclic assemblies are found, returns empty list.

        :param sequences: list of sequences
        :type sequences: list
        :param min_bases: minimum junction length
        :type min_bases: int
        :param max_bases: maximum junction length. Smaller values will result in faster search, but may miss valid
                            assemblies.
        :return: list of Assembly instances. Sequence can be accessed using `assembly.product`
        :rtype: list
        """
        G = cls.interaction_graph(
            sequences,
            bind_reverse_complement=True,
            min_bases=min_bases,
            max_bases=max_bases,
            only_ends=True,
        )
        linear_paths = cls.linear_paths(G)
        if not linear_paths:
            return []
        assemblies = []
        for path in linear_paths:
            assembly = cls._path_to_assembly(path, G)
            assemblies.append(assembly)
        return assemblies

    @classmethod
    def cyclic_assemblies(
        cls, sequences, min_bases=Sequence.DEFAULTS.MIN_ANNEAL_BASES, max_bases=None
    ):
        """
        Finds all unique, longest cyclic assemblies. If no cyclic assemblies found, returns empty list.

        :param sequences: list of sequences
        :type sequences: list
        :param min_bases: minimum junction length
        :type min_bases: int
        :param max_bases: maximum junction length. Smaller values will result in faster search, but may miss valid
                            assemblies.
        :type max_bases: int
        :return: list of Assembly instances. Sequence can be accessed using `assembly.product`
        :rtype: list
        """
        G = cls.interaction_graph(
            sequences,
            bind_reverse_complement=True,
            min_bases=min_bases,
            max_bases=max_bases,
            only_ends=True,
        )

        nodes = list(G.nodes)
        fwd = nodes[: len(sequences)]
        rev = nodes[len(sequences) :]
        fpriority = list(range(len(fwd)))
        rpriority = list(range(len(fwd), len(fwd) * 2))
        priority_rank = fpriority + rpriority[1:] + [rpriority[0]]
        node_priority = dict(zip(fwd + rev, priority_rank))

        cyclic_paths = cls.cyclic_paths(G)
        if not cyclic_paths:
            return []
        assemblies = []
        cyclic_paths = sorted(
            cyclic_paths, key=lambda x: min([node_priority[n] for n in x])
        )
        for path in cyclic_paths:
            ranks = [node_priority[n] for n in path]
            mn = ranks.index(min(ranks))
            cyclic_path = path[mn:] + path[:mn]
            assembly = cls._path_to_assembly(cyclic_path, G, cyclic=True)
            assemblies.append(assembly)
        return assemblies



import toml
from os.path import dirname, abspath, join
import json
from os import getenv
from os.path import isfile
import sys

here = dirname(abspath(__file__))
INPATH = join(here, "../../pyproject.toml")
OUTPATH = join(here, "version.json")


def pull_version():
    """Pulls version from pyproject.toml and moves to version JSON"""
    config = toml.load(INPATH)
    ver_data = dict(config["tool"]["poetry"])
    try:
        with open(OUTPATH, "r") as f:
            print(">> Previous Version:")
            print(f.read())
    except:
        pass

    with open(OUTPATH, "w") as f:
        print("<< New Version (path={}):".format(OUTPATH))
        print(json.dumps(ver_data, indent=2))
        json.dump(ver_data, f, indent=2)


def parse_version():
    """Parses version JSON"""
    if isfile(OUTPATH):
        with open(OUTPATH, "r") as f:
            ver = json.load(f)
        return ver
    else:
        return {}


if __name__ == "__main__":
    pull_version()


ver = parse_version()


def get_version():
    v = ver.get("version", None)
    print(v)
    return v


def get_name():
    name = ver.get("name", None)
    print(name)
    return name


def verify_ci():
    tag = getenv("CIRCLE_TAG")

    if tag != ver.get("version", None):
        info = "Git tag: {0} does not match the version of this app: {1}".format(
            tag, ver.get("version", None)
        )
        sys.exit(info)


__version__ = ver.get("version", None)
__title__ = ver.get("name")
__author__ = ver.get("authors")
__homepage__ = ver.get("homepage")
__repo__ = ver.get("repository")



from Bio import pairwise2, SeqIO
from Bio.Align.Applications import MafftCommandline
from jdna.viewer import SequenceViewer

from jdna.interface import ClassInterface, Interface
import tempfile


class AlignInterface(ClassInterface):
    def mafft(self, sequences, mafft_exe="/usr/local/bin/mafft", verbose=False):
        seqs = self._mafft_sequences(sequences, mafft_exe, verbose)
        viewer = self._mafft_viewer(seqs)
        return viewer

    def _mafft_sequences(
        self, sequences, mafft_exe="/usr/local/bin/mafft", verbose=False
    ):
        in_file = tempfile.mkstemp()[1]
        self._cls.IO.fasta(sequences, in_file)

        mafft_cline = MafftCommandline(mafft_exe, input=in_file)
        if verbose:
            print(mafft_cline)
        mafft_cline.auto = True
        mafft_cline.adjustdirection = True
        result = mafft_cline()
        if verbose:
            print(result[1])

        out_file = tempfile.mkstemp()[1]
        with open(out_file, "w") as f:
            f.write(result[0])
        return self._cls.IO.read_fasta(out_file)

    @staticmethod
    def _mafft_viewer(seqs):
        viewer = SequenceViewer(
            seqs,
            sequence_labels=["({})".format(i) for i in range(len(seqs))],
            apply_indices=list(range(len(seqs))),
            name="MAFFTA alignment",
        )
        viewer.metadata["Sequence Names"] = "\n\t" + "\n\t".join(
            ["{} - {}".format(i, s.name) for i, s in enumerate(seqs)]
        )
        return viewer

    def instance(self, instance):
        return AlignInstanceInterface(instance, self)


class AlignInstanceInterface(Interface):
    def pairwise(self, other):
        """
        Perform a pairwise alignment using BioPython.

        :param other: the other sequence
        :type other: Sequence | basestring
        :return: list of alignments (tuples)
        :rtype: list
        """
        alignments = pairwise2.align.globalxx(
            str(self._inst.upper(), str(other).upper())
        )
        return alignments

    def print_alignment(self, other, max=1):
        """
        Print an alignment with another sequence as a view object. Output will be similar
        to the following:

        .. code::

            > "Alignment" (170bp)


            0         G-GC---G---G----G-C-------------G-TG-----A-T----T---------T--T---ATGTTCATGGACGCCCGGGT
                      | ||   |   |    | |             | ||     | |    |         |  |   ||||||||||||||||||||
                      GAGCCACGCACGTCCCGGCATATTAACTCCAAGCTGGTTCTACTCGGCTGGGCGGGCGTGATTTTATGTTCATGGACGCCCGGGT

            85        ATCAAGGCAGCGGCTCACGCCTCTCCACGCGG--GACAG--GTGAAC--TATC--C-G-ACTAGG---TATCAA-----AG--AC
                      |||||||||||||||   |  | |    | ||  || ||  | |||   |||   | | |  |||   | || |     ||  |
                      ATCAAGGCAGCGGCT---G--T-T----G-GGCAGA-AGAAG-GAA-AATAT-ATCAGGA--AGGCCGT-TC-AGGTTTAGGGA-

        :param other: the other sequence
        :type other: Sequence | basestring
        :param max: maximum number of alignments to display (default=1)
        :type max: int
        :return: None
        :rtype: None
        """
        alignments = self.pairwise(other)
        for a in alignments[:max]:
            mid = ""
            for x1, x2 in zip(a[0], a[1]):
                if "-" not in [x1, x2]:
                    mid += "|"
                else:
                    mid += " "
            viewer = SequenceViewer([a[0], mid, a[1]], name="Alignment")
            viewer.print()

    def sanger_reads(
        self,
        abi_filepaths,
        format="abi",
        mafft_exe="/usr/local/bin/mafft",
        verbose=False,
    ):
        seqrecords = []
        for filepath in abi_filepaths:
            seqrecords.append(SeqIO.read(filepath, format=format))
        sequences = [self._inst.IO.from_seqrecord(seq) for seq in seqrecords]
        return self._class_interface.mafft([self._inst] + sequences, mafft_exe, verbose)

    def mafft(self, sequences, mafft_exe="/usr/local/bin/mafft", verbose=False):
        return self._class_interface.mafft([self._inst] + sequences, mafft_exe, verbose)



"""
Represent linear or circularized nucleotides
"""

import itertools
import re
from collections import defaultdict
from copy import copy
from enum import IntFlag

import primer3
from Bio import Restriction

from jdna.align import AlignInterface
from jdna.alphabet import AmbiguousDNA, UnambiguousDNA
from jdna.format import format_sequence

# from jdna.align import Align
from jdna.io import IOInterface
from jdna.linked_list import Node, DoubleLinkedList, LinkedListMatch
from jdna.utils import random_color
from jdna.viewer import SequenceViewer, ViewerAnnotationFlag, StringColumn


class SequenceFlags(IntFlag):
    """Constants/Flags for sequences."""

    FORWARD = 1
    REVERSE = -1
    TOP = 1
    BOTTOM = -1


class Feature(object):
    """An annotation for a sequence."""

    def __init__(self, name, type=None, strand=None, color=None):
        self.name = name
        if type is None:
            type = "misc"
        self.type = type
        if strand is None:
            strand = SequenceFlags.FORWARD
        self.strand = strand
        if color is None:
            color = random_color()
        self.color = color
        # self._nodes = set()

    def reverse(self):
        self.strand = -self.strand

    def __str__(self):
        return "<Feature name='{name}' type='{tp}' color='{color}'".format(
            name=self.name, tp=self.type, color=self.color
        )

    def __repr__(self):
        return str(self)

    def __copy__(self):
        return self.__class__(self.name, self.type, self.strand, self.color)

    # @property
    # def nodes(self):
    #     return self._nodes
    #
    # def segments(self):
    #     return Sequence.segments(self.nodes)
    # visited = set()
    # pairs = set()
    # stop = lambda x: x not in self._nodes
    # for n in self._nodes:
    #     if n not in visited:
    #         tail = n
    #         for tail in n.fwd(stop_criteria=stop):
    #             visited.add(tail)
    #         head = n
    #         for head in n.rev(stop_criteria=stop):
    #             visited.add(head)
    #         pairs.add((head, tail))
    # return pairs

    def is_multipart(self):
        if len(self.segments) > 1:
            return True
        return False

    def _bind(self, nodes):
        pass
        # for n in nodes:
        #     self._nodes.add(n)

    # def _unbind(self, nodes):
    #     for n in nodes:
    #         if n in self._nodes:
    #             self._nodes.remove(n)


class BindPos(LinkedListMatch):
    def __init__(
        self,
        template_bounds,
        query_bounds,
        template,
        query,
        direction,
        strand=SequenceFlags.TOP,
    ):
        """
        Makes a sequence binding position.

        :param template_bounds_list: list of 2 len tuples containing starts and ends from a template
        :type template_bounds_list: template DoubleLinkedList
        :param query_bounds_list: list of 2 len tuples containing starts and ends from a query
        :type query_bounds_list: query DoubleLinkedList
        :param template: the template
        :type template: DoubleLinkedList
        :param query: the query
        :type query: DoubleLinkedList
        :param direction: If SequenceFlags.FORWARD, the binding position indicates binding forward, to the bottom strand
                            of a dsDNA sequence.
        :type direction: int
        :param strand: If SequenceFlags.BOTTOM, then the query is assumed to be the reverse_complement of the original
                        query
        :type strand: int
        """
        super().__init__(template_bounds, query_bounds, template, query)
        self.direction = direction
        self.strand = strand

        if self.direction == SequenceFlags.REVERSE:
            self.anneal = query.copy_slice(*self.query_bounds[::-1])
            self.five_prime_overhang = query.new_slice(None, self.query_end.prev())
            self.three_prime_overhang = query.new_slice(self.query_start.next(), None)
        else:
            self.anneal = query.copy_slice(*self.query_bounds)
            self.five_prime_overhang = query.new_slice(None, self.query_start.prev())
            self.three_prime_overhang = query.new_slice(self.query_end.next(), None)
        # self.anneal = self.primer[query_span[0]:query_span[1]+1]
        # self.five_prime_overhang = self.primer[:query_span[0]]
        # self.three_prime_overhang = self.primer[query_span[1]+1:]

    # def innitialize(self):
    #     if self.direction == SequenceFlags.REVERSE:
    #         if self.anneal:
    #             self.anneal.reverse_complement()
    #         if self.five_prime_overhang:
    #             self.five_prime_overhang.reverse_complement()
    #         if self.three_prime_overhang:
    #             self.three_prime_overhang.reverse_complement()
    #             length = len(self.three_prime_overhang)
    #         else:
    #             length = 0
    #         self.three_prime_overhang, self.five_prime_overhang = self.five_prime_overhang, self.three_prime_overhang
    #         self.query_span = (self.query_span[0] + length, self.query_span[1] + length)

    @classmethod
    def from_match(
        cls, linked_list_match, template, query, direction, strand=SequenceFlags.TOP
    ):
        """
        Return a binding pos
        :param linked_list_match: the linked list match
        :type linked_list_match: LinkedListMatch
        :return:
        :rtype:
        """
        return cls(
            linked_list_match.template_bounds,
            linked_list_match.query_bounds,
            template,
            query,
            direction,
            strand=strand,
        )

    @property
    def template_anneal(self):
        if self.strand == SequenceFlags.FORWARD:
            return Sequence.new_slice(self.start, self.end)
        else:
            return Sequence.new_slice(self.start, self.end)

    @property
    def query_anneal(self):
        if self.direction == SequenceFlags.FORWARD:
            return Sequence.new_slice(self.query_start, self.query_end)
        else:
            return Sequence.new_slice(
                self.query_end, self.query_start
            ).reverse_complement()

    def __repr__(self):
        return "<{cls} span={span} direction='{direction}' strand='{strand}' 5'='{five}' anneal='{anneal}' 3'='{three}'>".format(
            cls=self.__class__.__name__,
            span=self.span,
            direction=self.direction,
            strand=self.strand,
            five=self.five_prime_overhang.__repr__(),
            three=self.three_prime_overhang.__repr__(),
            anneal=self.anneal.__repr__(),
        )


class Nucleotide(Node):
    """Represents a biological nucleotide. Serves a :class:`Node` in teh :class:`Sequence` object."""

    __slots__ = ["data", "__next", "__prev", "_features", "alphabet"]

    def __init__(self, base, alphabet=AmbiguousDNA):
        """
        Nucleotide constructor

        :param base: base as a single character string
        :type base: basestring
        """
        super(Nucleotide, self).__init__(base)
        self._features = set()
        self.alphabet = alphabet

    @classmethod
    def random(cls):
        """Generate a random sequence"""
        return cls(UnambiguousDNA.random())

    @property
    def base(self):
        return self.data

    def equivalent(self, other) -> bool:
        return self.alphabet.compare(self.base, other.base)

    def complementary(self, other) -> bool:
        return self.base.upper() == AmbiguousDNA[other.base].upper()

    def to_complement(self):
        self.data = AmbiguousDNA[self.data]

    def set_next(self, nucleotide):
        self.cut_next()
        super(Nucleotide, self).set_next(nucleotide)
        Nucleotide.fuse_features(self, nucleotide)

    def set_prev(self, nucleotide):
        self.cut_prev()
        super(Nucleotide, self).set_prev(nucleotide)
        Nucleotide.fuse_features(nucleotide, self)

    def cut_prev(self):
        return self._cut(cut_prev=True)

    def cut_next(self):
        return self._cut(cut_prev=False)

    def _cut(self, cut_prev=True):
        for f in self.features:
            self.split_features(split_prev=cut_prev)
        if cut_prev:
            nxt = super().cut_prev()
        else:
            nxt = super().cut_next()
        return nxt

    @property
    def features(self):
        return self._features

    def add_feature(self, feature):
        self.features.add(feature)
        return feature

    def remove_feature(self, feature):
        self.features.remove(feature)

    def feature_fwd(self, feature):
        stop = lambda x: feature not in x.features
        return self._propogate(lambda x: x.next(), stop_criteria=stop)

    def feature_rev(self, feature):
        stop = lambda x: feature not in x.features
        return self._propogate(lambda x: x.prev(), stop_criteria=stop)

    def replace_feature(self, old_feature, new_feature):
        self.features[new_feature] = self.features[old_feature]
        self.remove_feature(old_feature)

    def copy_features_from(self, other):
        for f in other.features:
            i = other.features[f]
            if f not in self.features:
                self.add_feature(f)
        self._remove_overlapping_features()

    def get_feature_span(self, feature):
        start = self.feature_rev(feature)[-1]
        end = self.feature_fwd(feature)[-1]
        return (start.features[feature], end.features[feature])

    # def update_feature_span(self, feature, delta_i):
    #     start = self.feature_rev(feature)[-1]
    #     for n in start.feature_fwd(feature):
    #         n.ffeatures[feature] += delta_i

    def _remove_overlapping_features(self):
        # type: () -> Nucleotide
        feature_pairs = itertools.combinations(list(self.features.keys()), 2)
        tobedel = set()
        for f1, f2 in feature_pairs:
            if f1.name == f2.name:
                tobedel.add(f2)
        for tob in tobedel:
            self.remove_feature(tob)

    @staticmethod
    def _default_fuse_condition(f1, f2):
        return f1.name == f2.name

    @classmethod
    def fuse_features(cls, n1, n2, fuse_condition=None):
        if n1 is None or n2 is None:
            return
        if fuse_condition is None:
            fuse_condition = cls._default_fuse_condition

        if not (n1.next() is n2 and n2.prev() is n1):
            raise Exception("Cannot fuse non-consecutive features")

        for f1 in set(n1.features):
            for f2 in set(n2.features):
                if f1 is not f2 and fuse_condition(f1, f2):
                    for n in n2.feature_fwd(f2):
                        n.add_feature(f1)
                        n.remove_feature(f2)

    #
    # @staticmethod
    # def fuse_features(n1, n2):
    #     if n1 is None:
    #         return
    #     if n2 is None:
    #         return
    #
    #     delset = set()
    #
    #     for f1 in n1.features:
    #         for f2 in n2.features:
    #             f1_pos = n1.features[f1]
    #             f2_pos = n2.features[f2]
    #             f1_copy = copy(f1)
    #             # same name & consecutive position
    #             if f1 is f2:
    #                 continue
    #             if f1.name == f2.name and f1_pos + 1 == f2_pos:
    #                 delset.add((f1, f2, f1_copy))
    #     for f1, f2, f1_copy in delset:
    #         for n in n1.feature_rev(f1):
    #             try:
    #                 n.replace_feature(f1, f1_copy)
    #             except KeyError:
    #                 pass
    #         for n in n2.feature_fwd(f2):
    #             try:
    #                 n.replace_feature(f2, f1_copy)
    #             except KeyError:
    #                 pass

    def split_features(self, split_prev=True):
        x1 = self.prev()
        x2 = self
        if not split_prev:
            # then split_next
            x1 = self
            x2 = next(self)
        # If at the end, no splitting is necessary
        if x1 is None or x2 is None:
            return
        for f in x1.features:
            # If this feature spans
            if f in x2.features:
                # Grab the sequences for the split feature
                frag1 = x1.feature_rev(f)
                frag2 = x2.feature_fwd(f)

                # check if its a cyclic feature
                if x2 in frag1:
                    continue
                if x1 in frag2:
                    continue

                # Make two copies of the feature
                f1 = copy(f)
                f2 = copy(f)

                # Swap original feature for copy
                for n in frag1:
                    n.replace_feature(f, f1)
                for n in frag2:
                    n.replace_feature(f, f2)

    def copy(self):
        copied = super(Nucleotide, self).copy()
        for f in self.features:
            copied.add_feature(f)
        return copied


class Sequence(DoubleLinkedList):
    """Represents a biological sequence as a double linked list. Can be annotated with features."""

    class DEFAULTS(object):
        """Sequence defaults"""

        MIN_ANNEAL_BASES = 13
        FOREGROUND_COLORS = ["blue", "red"]
        BACKGROUND_COLORS = None
        ALPHABET = AmbiguousDNA

    FORWARD = SequenceFlags.FORWARD
    REVERSE = SequenceFlags.REVERSE
    TOP = SequenceFlags.TOP
    BOTTOM = SequenceFlags.BOTTOM
    NODE_CLASS = Nucleotide
    counter = itertools.count()

    def __init__(
        self,
        sequence=None,
        first=None,
        name=None,
        description="",
        metadata=None,
        cyclic=False,
        alphabet=DEFAULTS.ALPHABET,
    ):
        """

        :param sequence: sequence string
        :type sequence: basestring
        :param first: optional first Nucleotide to use as the 'head' to this Sequence
        :type first: Nucleotide
        :param name: optional name of the sequence
        :type name: basestring
        :param description: optional description of the sequence
        :type description: basestring
        :param metadata: additional sequence metadata
        :type metadata: dict
        :param cyclic: whether to make the sequence circular
        :type cyclic: bool
        :param alphabet: the base pair alphabet of this sequence which used for complementary and comparisons
                        (default: AmbiguousDNA)
        :type alphabet: jdna.alphabet.Alphabet
        """

        self.alphabet = alphabet
        super(Sequence, self).__init__(data=sequence, first=first, cyclic=cyclic)
        if name is None:
            name = ""
        self.name = name
        self.description = description
        if metadata is None:
            metadata = dict()
        self.metadata = metadata
        self._global_id = next(Sequence.counter)
        self._io = self.IO.instance(self)
        self._align = self.Align.instance(self)
        if cyclic:
            self.cyclic = cyclic

    def new_node(self, data):
        return self.NODE_CLASS(data, alphabet=self.alphabet)

    @property
    def io(self):
        return self._io

    @property
    def align(self):
        return self._align

    @property
    def global_id(self):
        return self._global_id

    @classmethod
    def random(cls, length):
        """Generate a random sequence"""
        seq = ""
        for i in range(length):
            seq += UnambiguousDNA.random().upper()
        if seq == "":
            return cls.empty()
        return cls(sequence=seq)

    @property
    def features(self):
        """
        Returns set of features contained in sequence.

        :return: set of features in this sequence
        :rtype: set
        """
        features_set = set()
        for i, n in enumerate(self):
            features_set.update(n.features)
        return features_set

    def feature_positions(self, with_nodes=False):
        """
        Return a list of feature positions.

        :param with_nodes: if True, will return a tuple composed of a feature to position dictionary and a feature to
                            start and end node. If False, will just return a feature to position dictionary
        :type with_nodes: bool
        :return: feature positions dictionary OR tuple of feature positions dictionary and feature node dictionary
        :rtype: tuple
        """
        index = 0
        feature_pos = defaultdict(list)
        feature_nodes = defaultdict(list)
        l = len(self)
        for n in self:
            for f in n.features:
                if feature_pos[f] and feature_pos[f][-1][-1] + 1 == index:
                    feature_pos[f][-1][-1] = index
                    feature_nodes[f][-1][-1] = n
                else:
                    feature_pos[f].append([index, index])
                    feature_nodes[f].append([n, n])
            index += 1

        # capture features that span the origin
        if self.cyclic:
            for k in feature_pos:
                positions = feature_pos[k]
                nodes = feature_nodes[k]
                if len(nodes) > 1:
                    if positions[0][0] == 0 and positions[-1][-1] == l - 1:
                        nodes[0][0] = nodes[-1][0]
                        positions[0][0] = positions[-1][0]
                        nodes.pop()
                        positions.pop()

        if with_nodes:
            return feature_pos, feature_nodes
        return feature_pos

    def feature_nodes(self):
        return self.feature_positions(with_nodes=True)[-1]

    def add_feature(self, start, end, feature):
        """
        Add a feature to the start and end positions (inclusive)

        :param start: start
        :type start: int
        :param end: end (inclusive)
        :type end: int
        :param feature: the feature to add
        :type feature: Feature
        :return: the added feature
        :rtype: Feature
        """
        feature_nts = list(self.inclusive_range(start, end))
        if end and feature_nts[-1] is not self[end]:
            if not self.cyclic:
                raise IndexError(
                    "Cannot add feature to {} to linear dna with bounds {}".format(
                        (start, end), (0, len(self))
                    )
                )
            else:
                raise IndexError("Cannot add feature to {}".format((start, end)))
        for n in self.inclusive_range(start, end):
            n.add_feature(feature)
        return feature

    def add_multipart_feature(self, positions, feature):
        """
        Add a multi-part feature (i.e. a disjointed feature)

        :param positions: list of start and ends as tuples ([(1,100), (110,200)]
        :type positions: list
        :param feature: the feature to add
        :type feature: Feature
        :return: the added feature
        :rtype: Feature
        """
        for i, j in positions:
            self.add_feature(i, j, feature)
        return feature

    # def print_features(self):
    #     raise NotImplementedError()

    def find_feature_by_name(self, name):
        """
        Find features by name

        :param name: feature name
        :type name: basestring
        :return: list of features
        :rtype: list
        """
        found = []
        for feature in self.feature_positions():
            if feature.name == name:
                found.append(feature)
        return found

    def annotate(self, start, end, name, feature_type=None, color=None, strand=None):
        """
        Annotate a regions

        :param start: start
        :type start: int
        :param end: end (inclusive)
        :type end: end
        :param name: feature name
        :type name: basestring
        :param feature_type: feature type (default=misc)
        :type feature_type: basestring
        :param color: optional feature color
        :type color: basestring
        :return: new feature
        :rtype: Feature
        """
        return self.add_feature(
            start, end, Feature(name, feature_type, strand=strand, color=color)
        )

    def reverse(self):
        features_set = set()
        if self.is_empty():
            return self
        nodes = self.nodes
        for s in nodes:
            s.swap()
            features_set.update(s.features)
        for f in features_set:
            f.reverse()
        self.head = nodes[-1]
        return self

    def complement(self):
        """Complement the sequence in place"""
        if self.is_empty():
            return self
        curr = self.head
        visited = set()
        while curr and curr not in visited:
            visited.add(curr)
            curr.to_complement()
            curr = next(curr)
        return self

    def c(self):
        """Complement the sequence in place"""
        return self.complement()

    def reverse_complement(self):
        """Reverse complement the sequence in place"""
        self.reverse()
        self.complement()
        return self

    def rc(self):
        """Reverse complement the sequence in place"""
        return self.reverse_complement()

    def cut(self, i, cut_prev=True):
        fragments = super(Sequence, self).cut(i, cut_prev)
        fragments = [Sequence(first=f.head) for f in fragments]
        return fragments

    def __copy__(self):
        copied = super(Sequence, self).__copy__()
        copied._global_id = next(self.counter)
        copied.name = self.name
        # feature_positions = self.feature_positions()
        # for feature, positions in feature_positions.items():
        #     copied.add_multipart_feature(positions, copy(feature))
        return copied

    # def anneal_to_bottom_strand(self, other, min_bases=10):
    #     for match in self.find_iter(other,
    #                                 min_query_length=min_bases,
    #                                 direction=self.Direction.REVERSE, ):
    #         yield match
    #
    # def anneal_to_top_strand(self, other, min_bases=10):
    #     for match in self.find_iter(other,
    #                                 min_query_length=min_bases,
    #                                 protocol=lambda x, y: x.complementary(y)):
    #         yield match

    def anneal_forward(self, other, min_bases=DEFAULTS.MIN_ANNEAL_BASES, depth=None):
        """Anneal a sequence in the forward direction"""
        for match in self.find_iter(
            other,
            min_query_length=min_bases,
            direction=self.Direction.REVERSE,
            depth=depth,
        ):
            yield BindPos.from_match(
                match, self, other, direction=self.Direction.FORWARD
            )

    def anneal_reverse(self, other, min_bases=DEFAULTS.MIN_ANNEAL_BASES, depth=None):
        """Anneal a sequence in the reverse direction"""
        for match in self.find_iter(
            other,
            min_query_length=min_bases,
            direction=(1, -1),
            protocol=lambda x, y: x.complementary(y),
            depth=depth,
        ):
            yield BindPos.from_match(
                match, self, other, direction=self.Direction.REVERSE
            )

    def anneal(self, ssDNA, min_bases=DEFAULTS.MIN_ANNEAL_BASES, depth=None):
        """Simulate annealing a single stranded piece of DNA to a double_stranded template"""
        for match in self.anneal_forward(ssDNA, min_bases=min_bases, depth=depth):
            yield match
        for match in self.anneal_reverse(ssDNA, min_bases=min_bases, depth=depth):
            yield match

    def dsanneal(self, dsDNA, min_bases=DEFAULTS.MIN_ANNEAL_BASES, depth=None):
        """Simulate annealing a double stranded piece of DNA to a double_stranded template"""
        for binding in self.anneal(dsDNA, min_bases=min_bases, depth=depth):
            yield binding
        for binding in self.anneal(
            dsDNA.copy().reverse_complement(), min_bases=min_bases, depth=depth
        ):
            binding.strand = SequenceFlags.BOTTOM
            yield binding

    def format(self, width=75, spacer=""):
        return format_sequence(str(self), width=width, spacer=spacer)

    # def _print_alignment_helper(self, sequences, name=''):
    #     for a in

    @classmethod
    def _apply_features_to_view(cls, sequence, view):
        for feature, positions in sequence.feature_positions().items():
            for pos in positions:
                direction = None
                if feature.strand == SequenceFlags.FORWARD:
                    direction = ViewerAnnotationFlag.FORWARD
                elif feature.strand == SequenceFlags.REVERSE:
                    direction = ViewerAnnotationFlag.REVERSE
                view.annotate(
                    pos[0],
                    pos[1],
                    label=feature.name,
                    fill=direction,
                    background=feature.color,
                )

    def view_bindings(self, bindings, view=None):
        if view is None:
            view = self.view(complement=True)
        for b in bindings:
            anneal = b.anneal
            primer_sequence = b.five_prime_overhang + anneal + b.three_prime_overhang
            annotation = StringColumn(
                [
                    str(primer_sequence),
                    " " * len(b.five_prime_overhang)
                    + "|" * len(anneal)
                    + " " * len(b.three_prime_overhang),
                ]
            )
            if b.direction == Sequence.FORWARD:
                view.annotate(b.span[0], b.span[1], annotation)
            if b.direction == Sequence.REVERSE:
                view.annotate(b.span[0], b.span[1], annotation.flip()[::-1], top=False)
        return view

    def view(
        self,
        indent=10,
        width=85,
        spacer=None,
        complement=False,
        features=True,
        **kwargs
    ):
        """
        Create a :class:`SequenceViewer` instance from this sequence. Printing the view object with
        annotations and complement will produce an output similar to the following:

        .. code::


            > "Unnamed" (550bp)


                                                                        ----------------GFP----------------
                                                                        |<START
                                                                        ----      -----------RFP-----------
            0         CCCAGGACTAGCGACTTTCCGTAACGCGACCTAACACCGGCCGTTCCTTCGAGCCAGGCAAATGTTACGTCACTTCCTTAGATTT
                      GGGTCCTGATCGCTGAAAGGCATTGCGCTGGATTGTGGCCGGCAAGGAAGCTCGGTCCGTTTACAATGCAGTGAAGGAATCTAAA

                      ------GFP------
                      -----------------------------------------RFP-----------------------------------------
            85        TGAACAGCGCCGTACCCCGATATGATATTTAGATATATAGCAGTTACACTTGGGGTTGCTATGGACTTAGATCTGCTGTATGTTT
                      ACTTGTCGCGGCATGGGGCTATACTATAAATCTATATATCGTCAATGTGAACCCCAACGATACCTGAATCTAGACGACATACAAA

                      -----------------------------------------RFP-----------------------------------------
            170       TCTTACCTTCCGCATCAGGGGACAATTCGCCAGTAGAATTCAGTTTGTGCGTGAGAACATAAGATTGAATCCCACGCAGGCACAA
                      AGAATGGAAGGCGTAGTCCCCTGTTAAGCGGTCATCTTAAGTCAAACACGCACTCTTGTATTCTAACTTAGGGTGCGTCCGTGTT

                      ---------------------RFP----------------------
            255       GCAGGGCGGGCAGACTCTATAGGTCCTAAGACCCTGAGACTGCGTCCTCAAGATACAGGTTAACAATCCCCGTATGGAGCCGTTC
                      CGTCCCGCCCGTCTGAGATATCCAGGATTCTGGGACTCTGACGCAGGAGTTCTATGTCCAATTGTTAGGGGCATACCTCGGCAAG

            340       TTAGCATGACCCGACAGGTGGGCTTGGCTCGCGTAAGTTGAGTGTTGCAGATACCTGCTGCTGCGCGGTCTAGGGGGAATCGCCG
                      AATCGTACTGGGCTGTCCACCCGAACCGAGCGCATTCAACTCACAACGTCTATGGACGACGACGCGCCAGATCCCCCTTAGCGGC

            425       ATTTTGACGTAGGATCGGTAATGGGCAGTAAACCCGCAACTATTTTCAGCACCAGATGCAAGTTTCCCTAGAAAGCGTCATGGTT
                      TAAAACTGCATCCTAGCCATTACCCGTCATTTGGGCGTTGATAAAAGTCGTGGTCTACGTTCAAAGGGATCTTTCGCAGTACCAA

            510       TGCAATCTCCTTAGGTCACAGCAAACATAGCAGCCCCTGT
                      ACGTTAGAGGAATCCAGTGTCGTTTGTATCGTCGGGGACA

        :param indent: indent between left column and base pairs view windo
        :type indent: int
        :param width: width of the view window
        :type width: int
        :param spacer: string to intersperse between sequence rows (default is newline)
        :type spacer: basestring
        :param complement: whether to include the complementary strand in the view
        :type complement: bool
        :param features: whether to include annotations/features in the view instance
        :type features: bool
        :return: the viewer object
        :rtype: SequenceViewer
        """
        if indent is None:
            indent = 10

        if width is None:
            width = 85

        seqs = [self]
        colors = self.DEFAULTS.FOREGROUND_COLORS[0]
        if complement:
            seqs.append(self.copy().complement())
            colors = self.DEFAULTS.FOREGROUND_COLORS
        if spacer is None:
            if complement:
                spacer = "\n"
            else:
                spacer = ""
        viewer = SequenceViewer(
            seqs,
            name=self.name,
            description=self.description,
            indent=indent,
            width=width,
            spacer=spacer,
            foreground_colors=colors,
            **kwargs
        )
        viewer.metadata.update(self.metadata)
        if features:
            self._apply_features_to_view(self, viewer)
        return viewer

    def upper(self):
        copied = self
        for n in copied:
            n.data = n.data.upper()
        return copied

    def lower(self):
        copied = self
        for n in copied:
            n.data = n.data.lower()
        return copied

    def print(
        self,
        indent=None,
        width=None,
        spacer=None,
        complement=False,
        features=True,
        **kwargs
    ):
        """
         Create and print a :class:`SequenceViewer` instance from this sequence. Printing the view object
         with annotations and complement will produce an output similar to the following:

         .. code::


             > "Unnamed" (550bp)


                                                                         ----------------GFP----------------
                                                                         |<START
                                                                         ----      -----------RFP-----------
             0         CCCAGGACTAGCGACTTTCCGTAACGCGACCTAACACCGGCCGTTCCTTCGAGCCAGGCAAATGTTACGTCACTTCCTTAGATTT
                       GGGTCCTGATCGCTGAAAGGCATTGCGCTGGATTGTGGCCGGCAAGGAAGCTCGGTCCGTTTACAATGCAGTGAAGGAATCTAAA

                       ------GFP------
                       -----------------------------------------RFP-----------------------------------------
             85        TGAACAGCGCCGTACCCCGATATGATATTTAGATATATAGCAGTTACACTTGGGGTTGCTATGGACTTAGATCTGCTGTATGTTT
                       ACTTGTCGCGGCATGGGGCTATACTATAAATCTATATATCGTCAATGTGAACCCCAACGATACCTGAATCTAGACGACATACAAA

                       -----------------------------------------RFP-----------------------------------------
             170       TCTTACCTTCCGCATCAGGGGACAATTCGCCAGTAGAATTCAGTTTGTGCGTGAGAACATAAGATTGAATCCCACGCAGGCACAA
                       AGAATGGAAGGCGTAGTCCCCTGTTAAGCGGTCATCTTAAGTCAAACACGCACTCTTGTATTCTAACTTAGGGTGCGTCCGTGTT

                       ---------------------RFP----------------------
             255       GCAGGGCGGGCAGACTCTATAGGTCCTAAGACCCTGAGACTGCGTCCTCAAGATACAGGTTAACAATCCCCGTATGGAGCCGTTC
                       CGTCCCGCCCGTCTGAGATATCCAGGATTCTGGGACTCTGACGCAGGAGTTCTATGTCCAATTGTTAGGGGCATACCTCGGCAAG

             340       TTAGCATGACCCGACAGGTGGGCTTGGCTCGCGTAAGTTGAGTGTTGCAGATACCTGCTGCTGCGCGGTCTAGGGGGAATCGCCG
                       AATCGTACTGGGCTGTCCACCCGAACCGAGCGCATTCAACTCACAACGTCTATGGACGACGACGCGCCAGATCCCCCTTAGCGGC

             425       ATTTTGACGTAGGATCGGTAATGGGCAGTAAACCCGCAACTATTTTCAGCACCAGATGCAAGTTTCCCTAGAAAGCGTCATGGTT
                       TAAAACTGCATCCTAGCCATTACCCGTCATTTGGGCGTTGATAAAAGTCGTGGTCTACGTTCAAAGGGATCTTTCGCAGTACCAA

             510       TGCAATCTCCTTAGGTCACAGCAAACATAGCAGCCCCTGT
                       ACGTTAGAGGAATCCAGTGTCGTTTGTATCGTCGGGGACA

         :param indent: indent between left column and base pairs view windo
         :type indent: int
         :param width: width of the view window
         :type width: int
         :param spacer: string to intersperse between sequence rows (default is newline)
         :type spacer: basestring
         :param complement: whether to include the complementary strand in the view
         :type complement: bool
         :param include_annotations: whether to include annotations/features in the view instance
         :type include_annotations: bool
         :return: the viewer object
         :rtype: SequenceViewer
         """
        self.view(
            indent=indent,
            width=width,
            spacer=spacer,
            complement=complement,
            features=features,
            **kwargs
        ).print()

    def tm(self):
        """
        Calculate the Tm of this sequence using primer3 defaults

        :return: the tm of the sequence
        :rtype: float
        """
        return primer3.calcTm(str(self).upper())

    def json(self):
        """Print sequence to a json dictionary"""
        annotations = []
        for feature, positions in self.feature_positions().items():
            for start, end in positions:
                annotations.append(
                    {
                        "start": start,
                        "end": end + 1,
                        "name": feature.name,
                        "color": feature.color,
                        "type": feature.type,
                        "strand": feature.strand,
                    }
                )

        return {
            "name": self.name,
            "isCircular": self.cyclic,
            "length": len(self),
            "bases": str(self),
            "annotations": annotations,
        }

    @classmethod
    def load(cls, data):
        """Load a sequence from a json formatted dictionary"""
        sequence = cls(data["bases"], name=data["name"])
        sequence.cyclic = data["isCircular"]
        sequence.name = data["name"]
        sequence.description = data.get("description", None)
        for a in data["annotations"]:
            sequence.annotate(
                a["start"], a["end"] - 1, a["name"], a["type"], a["color"]
            )
        return sequence

    def _collect_cut_sites(self, enzyme_site, cut1=None, cut2=None):
        if hasattr(enzyme_site, "charac"):
            cut1 = enzyme_site.charac[0]
            cut2 = enzyme_site.charac[1]
            enzyme_site = enzyme_site.charac[4]

        if isinstance(enzyme_site, str):
            enzyme_site = Sequence(enzyme_site)
        cut_sites = []
        for match in self.find_iter(enzyme_site):
            cut_sites.append(match.span[0] + cut1)
            cut_sites.append(match.span[1] + cut2 + 1)
        return cut_sites

    def digest(self, enzymes, as_names=False):
        """
        Supply either a Bio.RestrictionSite or a tuple of (seq, cut1, cut2)

        e.g. ('GTTTAAAC', 4, -4)

        :param enzymes: either a Bio.RestrictionSite or a tuple of (seq, cut1, cut2)
        :type enzymes: list (of tuple|Bio.RestrictionSite)
        :return: list of sequences
        :rtype: list
        """
        cut_sites = []

        if not isinstance(enzymes, list):
            enzymes = [enzymes]
        if as_names:
            enzymes = [getattr(Restriction, name) for name in enzymes]
        for enzyme in enzymes:
            if isinstance(enzyme, tuple):
                cut_sites += self._collect_cut_sites(*enzyme)
            else:
                cut_sites += self._collect_cut_sites(enzyme)
        return self.cut(cut_sites)

    def __repr__(self):
        max_width = 30
        replace = "..."
        display = int((max_width - len(replace)) / 2.0)
        s = str(self)
        if len(s) > display * 2:
            # diff = display*2 - len(s)
            s = s[:display] + "..." + s[-display:]
        return "Sequence('{}')".format(s)


Sequence.IO = IOInterface(Sequence)
Sequence.Align = AlignInterface(Sequence)



from jdna.regions.exceptions import RegionError


class Context(object):
    """ Abstract sequence. Circular or linear. """

    DEFAULT_START_INDEX = 0

    def __init__(
        self, length, circular, name=None, id=None, start_index=DEFAULT_START_INDEX
    ):
        """
        Context constructor

        :param length: length of context
        :type length: int
        :param circular: topology of context; True for circular contexts
        :type circular: bool
        :param start_index: the starting index offset for this context (usually 0 or 1)
        :type start_index: int
        """
        self.name = name
        self.id = id
        self.__start_index = start_index
        self.__length = length
        self.__circular = circular

    @property
    def circular(self):
        """Whether this context is circular"""
        return self.__circular

    @property
    def length(self):
        """The length of this context"""
        return len(self)

    @property
    def start(self):
        """
        Start index of allowable positions on context sequence

        :return:
        :rtype:
        """
        return self.__start_index

    @property
    def end(self):
        """
        End index of allowable positions on context sequence

        :return:
        :rtype:
        """
        return self.length + self.start - 1

    def span(self, x, y):
        """
        Calculates the inclusive distance between two points given the context sequence from left to right.

        ::

        e.g.
              context:    |------|
              positions:   x  y
              distance:    1234
        e.g.
              context:    |------|
              positions:   y  x
              distance:   56  1234

        :param x: position 1
        :type x: int
        :param y: position 2
        :type y: int
        :return:
        :rtype: int
        """
        x = self.translate_pos(x)
        y = self.translate_pos(y)
        mx = max(x, y)
        mn = min(x, y)
        m = mx - mn - 1
        if x > y:
            if self.circular:
                return self.length - m
            else:
                return None
        return int(m + 2)

    def within_bounds(self, pos, inclusive=True):
        """
        Whether a position is withing the bounds of acceptable indices given the context sequence.

        :param pos: position
        :type pos: int
        :param inclusive: whether to be inclusive (or exclusive)
        :type inclusive: bool
        :return: True (within bound) or False (not within bounds)
        :rtype: bool
        """
        if inclusive:
            return self.start <= pos <= self.end
        else:
            return self.start < pos < self.end

    def translate_pos(self, pos):
        """
        Translates the index to an allowable index on a circular sequence context.
        Throws RegionError if context is linear and pos is outside of bounds.

        ::

            Context:  |-------|
            C_Index:  1.......9..11
            Pos:      11
            TransPos: 2

        :param pos:
        :type pos:
        :return:
        :rtype:
        """
        if self.circular:
            cleared = False
            while not cleared:
                cleared = True
                if pos > self.end:
                    pos = pos - self.length
                    cleared = False
                if pos < self.start:
                    pos = pos + self.length
                    cleared = False
        else:
            if not self.within_bounds(pos, inclusive=True):
                raise RegionError(
                    "Position {} outside of bounds for linear region [{} {}].".format(
                        pos, self.start, self.end
                    )
                )
        return pos

    def __eq__(self, other):
        """Whether another context is functionally equivalent"""
        return (
            self.circular == other.circular
            and self.start == other.start
            and self.end == other.end
            and self.length == other.length
        )

    def __len__(self):
        """The length of the context"""
        return self.__length

    def __str__(self):
        return "Context(length={length}, circular={circular}, start_index={start_index})".format(
            length=self.length, circular=self.circular, start_index=self.start
        )

    def __repr__(self):
        return str(self)



"""
Linked list model to represent linear or circular sequences
"""

from copy import copy
from enum import IntFlag
from functools import reduce


class LinkedListException(Exception):
    """Generic linked list exception"""


class LinkedListIndexError(LinkedListException, IndexError):
    """Indices were out of bounds for LinkedList"""


class Node(object):
    """
    A node in a linked list
    """

    __slots__ = ["data", "__next", "__prev"]

    def __init__(self, data):
        """
        Node constructor. Stores a single piece of data.

        :param data: data
        :type data: any
        """
        self.data = data
        self.__next = None
        self.__prev = None

    def prev(self):
        """
        Return the previous node

        :return: the previous node
        :rtype: Node
        """
        return self.__prev

    def next(self):
        """
        Return the next node

        :return: the next node
        :rtype: Node
        """
        return self.__next

    def add_next(self, data):
        """
        Create a new node and add to next

        :param data: any data
        :type data: any
        :return: the new node
        :rtype: Node
        """
        new_node = self.__class__(data)
        self.set_next(new_node)
        return new_node

    def add_prev(self, data):
        """
        Create a new node and add to previous

        :param data: any data
        :type data: any
        :return: the new node
        :rtype: Node
        """
        new_node = self.__class__(data)
        self.set_prev(new_node)
        return new_node

    def cut_next(self):
        """
        Cut the next node, return the cut node.

        :return: the cut (next) node
        :rtype: Node
        """
        next_node = next(self)
        if next_node is not None:
            next_node.__prev = None
        self.__next = None
        return next_node

    def cut_prev(self):
        """
        Cut the previous node, return the cut node.

        :return: the cut (previous) node
        :rtype: Node
        """
        prev_node = self.prev()
        if prev_node is not None:
            prev_node.__next = None
        self.__prev = None
        return prev_node

    def _break_connections(self):
        """
        Break connections in this node.
        :return:
        :rtype:
        """
        self.set_next(None)
        self.set_prev(None)

    def remove(self):
        """
        Remove node from linked list, connecting the previous and next nodes together.

        :return: None
        :rtype: None
        """
        next_node = next(self)
        prev_node = self.prev()
        if next_node is not None:
            next_node.set_prev(prev_node)
        if prev_node is not None:
            prev_node.set_next(next_node)
        self._break_connections()
        return

    def swap(self):
        """
        Swap the previous and next nodes.

        :return: None
        :rtype: None
        """
        temp = self.__next
        self.__next = self.__prev
        self.__prev = temp

    def set_next(self, node):
        """
        Set the next node

        :param node:
        :type node:
        :return:
        :rtype:
        """
        if node is not None:
            node.__prev = self
        self.__next = node

    def set_prev(self, node):
        """
        Set the previous node

        :param node:
        :type node:
        :return:
        :rtype:
        """
        if node is not None:
            node.__next = self
        self.__prev = node

    def has_next(self):
        return self.__next is not None

    def has_prev(self):
        return self.__prev is not None

    def _propogate(self, next_method, stop=None, stop_criteria=None):
        visited = set()
        curr = self
        while True:
            if (
                curr is None
                or curr in visited
                or (stop_criteria and stop_criteria(curr))
            ):
                return
            yield curr
            if curr is stop:
                return
            visited.add(curr)
            curr = next_method(curr)

    def fwd(self, stop_node=None, stop_criteria=None):
        """
        Propogates forwards until stop node is visited or stop criteria is reached.

        :param stop_node:
        :type stop_node:
        :param stop_criteria:
        :type stop_criteria:
        :return:
        :rtype:
        """
        return self._propogate(
            lambda x: x.next(), stop=stop_node, stop_criteria=stop_criteria
        )

    def rev(self, stop_node=None, stop_criteria=None):
        """
        Propogates backwards until stop node is visited or stop criteria is reached.

        :param stop_node:
        :type stop_node:
        :param stop_criteria:
        :type stop_criteria:
        :return:
        :rtype:
        """
        return self._propogate(
            lambda x: x.prev(), stop=stop_node, stop_criteria=stop_criteria
        )

    def find_first(self):
        """
        Find the head node

        :return:
        :rtype:
        """
        rev = self.rev()
        first = self
        for first in rev:
            pass
        return first

    def find_last(self):
        """
        Find the tail node

        :return:
        :rtype:
        """
        fwd = self.fwd()
        last = self
        for last in fwd:
            pass
        return last

    def longest_match(self, node, next_method=None):
        """
        Find the longest match between two linked_lists

        :param node: the node to compare
        :type node: Node
        :param next_method: how to obtain the next node
        :type next_method: callable
        :return: list of tuples containing matching nodes
        :rtype: list
        """
        if next_method is None:
            next_method = lambda x: x.next()
        x1 = self
        x2 = node
        start = None
        end = None
        while x1 and x2 and x1.equivalent(x2):
            if start is None:
                start = (x1, x2)
            end = (x1, x2)
            x1 = next_method(x1)
            x2 = next_method(x2)
        return start, end

    def _complete_match(self, node, next_method):
        """
        Return whether the longest match between two nodes is equivalent.

        :param node: the node to compare
        :type node: Node
        :param next_method: how to obtain the next node
        :type next_method: callable
        :return: whether the longest match between two nodes is equivalent
        :rtype: bool
        """
        l = self.longest_match(node, next_method)
        if not l:
            return False
        t1, t2 = l[-1]
        return not (next_method(t1) and next_method(t2))

    def complete_match_fwd(self, y):
        return self._complete_match(y, lambda x: next(x))

    def complete_match_rev(self, y):
        return self._complete_match(y, lambda x: x.prev())

    def equivalent(self, other):
        """Evaluates whether two nodes hold the same data"""
        return self.data == other.data

    def __next__(self):
        return self.next()

    def copy(self):
        return self.__class__(self.data)

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        raise NotImplementedError(
            "copy.deepcopy not implemented with class"
            "{}. Use copy.copy instead.".format(self.__class__.__name__)
        )

    def __repr__(self):
        return str(self)

    def __str__(self):
        return str(self.data)


class EmptyNode(Node):
    def __init__(self):
        pass

    # @property
    # def __next(self):
    #     return EmptyNode()
    #
    # @property
    # def __prev(self):
    #     return EmptyNode()

    def __eq__(self, other):
        return False

    def _propogate(self, *args, **kwargs):
        return
        yield

    def __str__(self):
        return ""


class LinkedListMatch(object):
    """
    A match object
    """

    def __init__(self, template_bounds, query_bounds, template, query):
        self.span = tuple(template.indices_of(template_bounds))
        self.query_span = tuple(query.indices_of(query_bounds))
        self.query_bounds = query_bounds
        self.template_bounds = template_bounds

    @classmethod
    def batch_create(cls, template_bounds_list, query_bounds_list, template, query):
        """
        Efficiently create several LinkedListMatches from lists of template starts/ends and query starts/ends

        :param template_bounds_list: list of 2 len tuples containing starts and ends from a template
        :type template_bounds_list: template DoubleLinkedList
        :param query_bounds_list: list of 2 len tuples containing starts and ends from a query
        :type query_bounds_list: query DoubleLinkedList
        :param template: the template
        :type template: DoubleLinkedList
        :param query: the query
        :type query: DoubleLinkedList
        :return: matchese
        :rtype: list of LinkedListMatch
        """
        if len(template_bounds_list) != len(query_bounds_list):
            raise LinkedListIndexError(
                "Cannot create Matches The template bounds list must be same size as query_"
                "bounds list"
            )
        if not template_bounds_list:
            return []
        template_nodes = reduce(lambda x, y: list(x) + list(y), template_bounds_list)
        query_nodes = reduce(lambda x, y: list(x) + list(y), query_bounds_list)
        template_indices = template.indices_of(template_nodes)
        query_indices = query.indices_of(query_nodes)

        matches = []
        for i in range(0, len(template_nodes), 2):
            new_match = cls.__new__(cls)
            new_match.span = (template_indices[i], template_indices[i + 1])
            new_match.query_span = (query_indices[i], query_indices[i + 1])
            new_match.query_bounds = (query_nodes[i], query_nodes[i + 1])
            new_match.template_bounds = (template_nodes[i], template_nodes[i + 1])
            matches.append(new_match)
        return matches

    @property
    def start(self):
        return self.template_bounds[0]

    @property
    def end(self):
        return self.template_bounds[1]

    @property
    def query_start(self):
        return self.query_bounds[0]

    @property
    def query_end(self):
        return self.query_bounds[1]

    def __repr__(self):
        return "<{cls} span={span}, qspan={query_span}>".format(
            cls=self.__class__.__name__, span=self.span, query_span=self.query_span
        )

    def __str__(self):
        return self.__repr__()


class DoubleLinkedList(object):
    """
    A generic double linked list class.
    """

    class Direction(IntFlag):
        FORWARD = 1
        REVERSE = -1

    NODE_CLASS = Node

    def __init__(self, data=None, first=None, cyclic=False):
        """
        linked list construction

        :param data: iterable data
        :type data: iterable
        :param first: first node
        :type first: Node
        """
        self._head = EmptyNode()
        if data is not None:
            self.initialize(data)
        elif first is not None:
            self._head = first
        if cyclic:
            self.circularize()

    def new_node(self, data):
        return self.NODE_CLASS(data)

    def initialize(self, sequence):
        prev = None
        for i, d in enumerate(sequence):
            curr = self.new_node(d)
            if i == 0:
                self._head = curr
            if prev:
                prev.set_next(curr)
            prev = curr

    def is_empty(self):
        return isinstance(self._head, EmptyNode)

    @property
    def head(self):
        return self._head
        if self.is_empty():
            return self._head
        if self.cyclic:
            return self._head
        first = self._head.find_first()
        self._head = first
        return self._head

    @head.setter
    def head(self, new_head):
        self._head = new_head
        return new_head

    @property
    def tail(self):
        return self.head.find_last()

    # TODO: This method is inefficient, but can probably be managed more manually (i.e. anytime a manipulation occurs)
    @property
    def cyclic(self):
        if self.is_empty():
            return False
        visited = set()
        curr = self._head
        while curr:
            if curr in visited:
                return True
            visited.add(curr)
            curr = next(curr)
        return False

    @cyclic.setter
    def cyclic(self, b):
        if self.cyclic and not b:
            return self.linearize()
        elif not self.cyclic and b:
            return self.circularize()

    @property
    def circular(self):
        """Alias for cyclic"""
        return self.cyclic

    @circular.setter
    def circular(self, v):
        """Alias for cyclic"""
        self.cyclic = v

    def circularize(self):
        if not self.cyclic:
            self.tail.set_next(self.head)
        return self

    def linearize(self, i=0):
        this_i = self.get(i)
        this_i.cut_prev()
        self.head = this_i
        return self

    @property
    def nodes(self):
        n = list(self.head.fwd())
        return n

    def get(self, i):
        if i is None:
            return None
        elif i < 0:
            return self.nodes[i]
        for index, n in enumerate(self):
            if index == i:
                return n
        raise LinkedListIndexError(
            "There is no node at index '{}'. There are {} nodes.".format(i, len(self))
        )

    def cut(self, i, cut_prev=True):
        if isinstance(i, tuple):
            i = list(i)
        if isinstance(i, int):
            i = [i]
        # Special case in which i == len
        i = list(set(i))
        if len(self) in i and cut_prev:
            i.remove(len(self))
            if self.cyclic:
                i.append(0)
        i = list(set(i))
        i.sort()
        self._check_if_in_bounds(i)
        self_copy = copy(self)
        all_nodes = self_copy.nodes
        cut_nodes = []
        for cut_loc in i:
            node = all_nodes[cut_loc]
            if cut_prev:
                c = node.cut_prev()
                if c is not None:
                    cut_nodes.append(c)
                cut_nodes.append(node)
            else:
                cut_nodes.append(node)
                c = node.cut_next()
                if c is not None:
                    cut_nodes.append(c)
        return self.segments(cut_nodes)

    @staticmethod
    def collect_nodes(nodes):
        """
        Return all visisted nodes and return an unordered set of nodes.

        :return: all visited nodes
        :rtype: set
        """
        visited = set()
        for n in nodes:
            if n not in visited:
                for tail in n.fwd(stop_criteria=lambda x: x in visited):
                    visited.add(tail)
                for head in n.rev(stop_criteria=lambda x: x in visited):
                    visited.add(head)
        return visited

    def all_nodes(self):
        return self.collect_nodes([self.head])

    @staticmethod
    def find_ends(nodes):
        """Efficiently finds the head and tails from a group of nodes."""
        visited = set()
        heads = []
        tails = []
        for n in nodes:
            if n not in visited:
                head = n
                tail = n
                visited_tails = set()
                visited_heads = set()
                for tail in n.fwd(stop_criteria=lambda x: x in visited):
                    visited_tails.add(tail)
                for head in n.rev(stop_criteria=lambda x: x in visited):
                    visited_heads.add(head)
                visited = visited.union(visited_heads)
                visited = visited.union(visited_tails)
                if head not in heads and tails not in tails:
                    if not head.has_prev() is None and not tail.has_next():
                        heads.append(head)
                        tails.append(tail)
        return zip(heads, tails)

    @classmethod
    def segments(cls, nodes):
        return [cls(first=h) for h, _ in cls.find_ends(nodes)]

    def insert(self, node_list, i, copy_insertion=True):
        if i == len(self):
            pass
        else:
            self._check_if_in_bounds(i)
        if node_list.cyclic:
            raise TypeError("Cannot insert a cyclic sequence")
        if copy_insertion:
            node_list = copy(node_list)
        # TODO: This copies the insertion sequence, you want that?
        if i == len(self):
            loc2 = None
            loc1 = self.get(i - 1)
        else:
            loc2 = self.get(i)
            loc1 = loc2.prev()
        first = node_list.nodes[0]
        last = node_list.nodes[-1]
        first.set_prev(loc1)
        last.set_next(loc2)
        if (
            i == 0
        ):  # Special case in which user inserts sequence in front of their sequence; they probably intend to re-index it
            self.head = first
        return self

    def remove(self, i):
        self._check_if_in_bounds(i)
        to_be_removed = self.get(i)
        new_first = self.head
        if i == 0:
            new_first = next(new_first)
        to_be_removed.remove()
        self.head = new_first
        return

    def reindex(self, i):
        self._check_if_in_bounds(i)
        if not self.cyclic:
            raise TypeError(
                "Cannot re-index a linear {}".format(self.__class__.__name__)
            )
        self.head = self.get(i)
        return self

    def _check_if_in_bounds(self, num):
        if isinstance(num, int):
            num = [num]
        for n in num:
            mn = 0
            mx = len(self) - 1
            if n < 0 or n > mx:
                raise LinkedListIndexError(
                    "Index {} out of acceptable bounds ({}, {})".format(n, mn, mx)
                )

    # TODO: implement yield in find_iter, search_all should call this
    # TODO: query should be any interable
    # TODO: [:1] and [-10:] style cuts should be available
    # TODO: documentation for methods
    # TODO: move DoubleLinkedList to its own thing?
    # TODO: element insertion
    # TODO: search should return a 'cut' of the sequence
    def search_all(self, query):
        curr_node = self.head
        q_node = query.head
        i = 0
        found = []
        visited = set()
        while curr_node and curr_node not in visited:
            visited.add(curr_node)
            if curr_node.complete_match_fwd(q_node):
                found.append((i, curr_node))
            curr_node = next(curr_node)
            i += 1
        return found

    def longest_match(self, other):
        n1 = self.head
        n2 = other.head
        start = None
        while n1 and n2 and n1.equivalent(n2):
            if start is None:
                start = (n1, n2)
            n1 = next(n1)
            n2 = next(n2)

            curr = next(curr)
        for n in self:
            if other_node.equivalent(n):
                stop = n
                if start is None:
                    start = stop
            other_node = next(other_node)

    # def longest_match(self, other):
    #     matched = []
    #     i = 0
    #     j = i
    #     for x1, x2 in zip(self, other):
    #         if x1.equivalent(x2):
    #             j += 1
    #             matched.append((x1, x2))
    #     if j > len(self):
    #         if not self.cyclic:
    #             raise Exception("Template was expected to be cyclic but was not")
    #         j -= len(self)
    #     return LinkedListMatch(self.head, )
    #     if len(matched) == qlen:
    #         if j > len(self):
    #             if self.cyclic:
    #                 j -= len(self)
    #             else:
    #                 raise Exception("Template was expected to be cyclic but was not")
    #         yield LinkedListMatch(curr_node, x1, span=(i, j - 1))

    @classmethod
    def match(
        cls,
        n1,
        n2,
        query_direction=Direction.FORWARD,
        template_direction=Direction.FORWARD,
        protocol=None,
    ):
        """"""
        if protocol is None:
            protocol = lambda x, y: x.equivalent(y)

        iterators = {
            cls.Direction.FORWARD: lambda x: x.fwd(),
            cls.Direction.REVERSE: lambda x: x.rev(),
        }
        for x1, x2 in zip(
            iterators[template_direction](n1), iterators[query_direction](n2)
        ):
            if protocol(x1, x2):
                yield (x1, x2)
            else:
                return

    def find_iter(
        self,
        query,
        min_query_length=None,
        direction=Direction.FORWARD,
        protocol=None,
        depth=None,
    ):
        """
        Iteratively finds positions that match the query.

        :param query: query list to find
        :type query: DoubleLinkedList
        :param min_query_length: the minimum number of matches to return. If None, find_iter will only return complete
                                matches
        :type min_query_length: inst
        :param direction: If Direction.FORWARD (+1), find iter will search from the query head and search forward,
                        potentially leaving a 'tail' overhang on the query. If
                        Direction.REVERSE (-1), find iter will search from the query tail and search reverse, potentially
                        leaving a 'head' overhang on the query. If a tuple, the template_direction and query_direction are
                        set respectively.
        :type direction: int or tuple
        :param protocol: the callable taking two parameters (as Node) to compare during find. If None, defaults
                        to 'equivalent'
        :type protocol: callable
        :return: list of LinkedListMatches
        :rtype: list
        """
        if isinstance(direction, tuple):
            template_direction, query_direction = direction
        else:
            template_direction, query_direction = direction, direction

        curr_node = self.head
        visited = set()
        if self.Direction.REVERSE == query_direction:
            query_start = query.tail
        else:
            query_start = query.head
        qlen = len(query)
        if min_query_length is None:
            min_query_length = qlen

        template_nodes = []
        query_nodes = []
        index = 0
        while curr_node and curr_node not in visited:
            index += 1
            if depth is not None and index > depth:
                break
            visited.add(curr_node)
            matches = list(
                self.match(
                    curr_node,
                    query_start,
                    query_direction=query_direction,
                    template_direction=template_direction,
                    protocol=protocol,
                )
            )
            if self.Direction.REVERSE == template_direction:
                matches = matches[::-1]
            if len(matches) >= min_query_length:
                query_nodes.append([matches[0][1], matches[-1][1]])
                template_nodes.append([matches[0][0], matches[-1][0]])
            curr_node = next(curr_node)
        return LinkedListMatch.batch_create(template_nodes, query_nodes, self, query)

    def reverse(self):
        if self.is_empty():
            return self
        nodes = self.nodes
        for s in nodes:
            s.swap()
        self.head = nodes[-1]
        # if self.cyclic:
        #     self.reindex(1)
        return self

    def fuse_in_place(self, seq):
        if seq.is_empty():
            return self
        elif self.is_empty():
            return seq.copy()
        f = self.head.find_last()
        l = seq.head
        f.set_next(l)
        return self

    def fuse(self, other):
        return copy(self).fuse_in_place(copy(other))

    def copy(self):
        return self.__copy__()

    @classmethod
    def empty(cls):
        return cls()

    @staticmethod
    def empty_iterator():
        return
        yield

    def node_range(self, start, end):
        """Iterate between 'start' to 'end' nodes (inclusive)"""
        for n in start.fwd(stop_node=end):
            yield n

    @classmethod
    def new_slice(cls, start, end):
        """Return a copy of the sequence between 'start' and 'end' nodes)"""
        nodes = []
        if start is not None:
            for n in start.fwd(stop_node=end):
                nodes.append(n)
                # new_node = copy(n)
                # new_nodes.append(new_node)
                # new_node.set_prev(prev)
                # prev = new_node
        else:
            if end is None:
                return cls.empty()
            for n in end.rev(stop_node=start):
                nodes.append(n)
            nodes = nodes[::-1]
        if end is not None and nodes[-1] is not end:
            return cls()
        if start is not None and nodes[0] is not start:
            return cls()

        next = None
        for n in nodes[::-1]:
            n = copy(n)
            n.set_next(next)
            next = n
        n.set_prev(None)

        return cls(first=n)

    def copy_slice(self, start, end):
        """Return a copy of the sequence between 'start' and 'end' nodes. If start is None,
        return the slice copy from head to end. If end is None, return copy from start to tail.
        If both start and end are None return None."""
        if start is None:
            start = self.head
        return self.new_slice(start, end)

    def range(self, i, j):
        """Returns an iterator from node at 'i' to node at 'j-1'"""
        if i == j:
            return self.empty_iterator()
        if j is not None:
            if j == 0 and not self.cyclic:
                return self.empty_iterator()
            j = j - 1
        try:
            return self.inclusive_range(i, j)
        except LinkedListIndexError:
            return self.empty_iterator()

    def inclusive_range(self, i, j):
        """Return generator for inclusive nodes between index i and j. If i is None, assume i is the head node."""
        start = self.get(i)
        if start is None:
            start = self.head
        try:
            stop = self.get(j)
        except LinkedListIndexError:
            stop = None
        stop_hit = False
        for n in start.fwd(stop_node=stop):
            yield n
            if n is stop:
                stop_hit = True
        if stop is not None and not stop_hit:
            raise LinkedListIndexError(
                "Inclusive indices {} out of bounds".format((i, j))
            )

    def index_of(self, node):
        for i, n in enumerate(self):
            if n is node:
                return i

    def indices_of(self, nodes):
        index_dict = {}
        for i, n in enumerate(self):
            for node in nodes:
                if n is node:
                    index_dict[n] = i
        return [index_dict.get(n, None) for n in nodes]

    def data(self):
        return [n.data for n in self]

    def compare(self, other):
        """
        Compares two linked lists. If both are cyclic, will attempt to reindex

        :param other: other linked list
        :type other: DoubleLinkedList
        :return: whether sequence data is equivalent
        :rtype: bool
        """
        if self.cyclic and other.cyclic:
            anchors = []
            if len(other) > 100:
                for a in self.find_iter(other[:20]):
                    anchors.append(a.span[0])
            else:
                anchors = range(len(other))
            for a in anchors:
                temp = self.copy().reindex(a)
                if temp == other:
                    return True
            return False
        else:
            return self == other

    def left_trim(self, i):
        if self.cyclic:
            raise IndexError("Cannot chop a cyclic sequence.")
        return self.cut(i)[-1]

    def right_trim(self, i):
        if self.cyclic:
            raise IndexError("Cannot chop a cyclic sequence.")
        return self.cut(i, cut_prev=False)[0]

    def __eq__(self, other):
        if self.cyclic != other.cyclic:
            return False
        data1 = [n.data for n in self]
        data2 = [n.data for n in other]
        return data1 == data2

    def __add__(self, other):
        return self.fuse(other)

    def __getitem__(self, key):
        if isinstance(key, slice):
            if key.step and key.step > 1:
                raise LinkedListIndexError(
                    "Step > 1 is not supported for sliced object of '{}'".format(
                        self.__class__.__name__
                    )
                )

            if key.start is not None and key.stop is not None:
                if key.start == key.stop:
                    return self.empty()

            new_list = self.__copy__()
            if key.start is None and key.stop is None:
                if key.step == -1:
                    return new_list.reverse()
                else:
                    return new_list

            try:
                new_nodes = list(new_list.range(key.start, key.stop))
            except LinkedListIndexError:
                return self.empty()
            if not new_nodes:
                return self.empty()
            start = new_nodes[0]
            end = new_nodes[-1]

            start.cut_prev()
            end.cut_next()
            return self.__class__(first=start)
        return self.get(key)

    def __contains__(self, item):
        return item in self.all_nodes()

    def __copy__(self):
        head = None
        prev = None
        for n in self.nodes:
            new_node = n.copy()
            if head is None:
                head = new_node
            new_node.set_prev(prev)
            prev = new_node
        copied = type(self)(first=head)
        if self.cyclic:
            copied.cyclic = self.cyclic
        return copied

    def __deepcopy__(self, memo):
        raise NotImplementedError(
            "copy.deepcopy not implemented with class"
            "{}. Use copy.copy instead.".format(self.__class__.__name__)
        )

    def __reversed__(self):
        copied = self.copy()
        copied.reverse()
        return copied

    def __len__(self):
        length = 0
        for _ in self:
            length += 1
        return length

    def __iter__(self):
        return self.head.fwd()

    def __repr__(self):
        return "<{cls} data='{data}'>".format(
            cls=self.__class__.__name__, data=str(self)
        )
        return str(self)

    def __str__(self):
        return "".join(str(x) for x in self)



"""
Project: jdna
File: region
Author: Justin
Date: 2/21/17

Description: Basic functionality for defining regions of linear, circularized, or reversed
regions of a sequence.

"""
from jdna.regions.exceptions import RegionError
from jdna.regions.context import Context


def force_same_context(error=False):
    """ Wrapper that returns False or raises Error if other Region has a different context. If error==False,
    then wrapped function returns False. If error=True, then raises a RegionError. """

    def context_wrapper(fxn):
        def check_context(*args, **kwargs):
            self = args[0]
            other = args[1]
            if not self.same_context(other):
                if error:
                    raise RegionError(
                        "Cannot compare two regions if they have different sequence contexts."
                    )
                else:
                    return False
            return fxn(*args, **kwargs)

        return check_context

    return context_wrapper


class Region(object):
    """
    Classifies an abstract region of a sequence. A region is defined by the inclusive "start" and "end"
    positions in context of an arbitrary sequence defined by the start_index and length.

    Regions can be circular or linear. For circular regions, negative indicies and indicies greater
    than the length are allowable and will be converted to appropriate indices.

    Direction of the region can be FORWARD or REVERSE or BOTH. For reversed directions, start
    and end positions should be flipped.

    Alternative start_index can be used (DEFAULT: 1) to handle sequences that start at 0 or 1.

    A new Region can be created either by defining the start and end positions or defining the
    start position and length by Region.create(length, circular) ::

        E.g. Linear Region
            length: 9
            start_index: 1
            start: 2
            end: 5
            length = 5-2+1 = 4
            Context:  |-------|
            C_Index:  1.......9
            Region:    2..4

        E.g. Circular Region
            length: 9
            start_index: 1
            start: 8
            end: 2
            length = 4
            Context:  |-------|
            C_Index:  1.......9
            Region:   .2     8.

        E.g. Calling Region(5, 7, 9, True, direction=REVERSE, start_index=1)
                  |-------|         Context
                  123456789
                  <<<<| |<<        Region Direction
                      s e          Start (s) and end (e)
    """

    START_INDEX = 0
    FORWARD = 1
    REVERSE = -1
    BOTH = 2

    def __init__(self, start, end, context, direction=FORWARD, name=""):
        """
        Annotates some region from a contextual sequence with either circular or linear topologies.

        Making a forward region:
            Region(1, 5, context=Context(length=10, circular=False, start_index=1))
            
        Making a reverse region:
            Region(5, 1, direction=Region.REVERSE, context=Context(length=10, circular=False, start_index=1))

        Invalid regions:
            Region(1, 5, direction=Region.REVERSE, context=Context(length=10, circular=False, start_index=1))
                > This would imply a circular sequence context since the region starts at 1 goes in reverse through an
                origin to index 5.

            Region(5, 1, direction=Region.FORWARD, context=Context(length=10, circular=False, start_index=1))
                This would imply a circular sequence context since the region starts at 5 goes forward through an
                origin to index 1.

        :param start: start index of region
        :type start: int
        :param end: end index of region
        :type end: int
        :param context: context sequence for this region
        :type context: Context
        :param direction: Region.FORWARD or Region.REVERSE
        :type direction: int
        :param name: optional name of region
        :type name: str
        """

        self.name = name
        self.context = context
        self.__start = None
        self.__end = None
        self.start = start
        self.end = end
        if direction == Region.REVERSE:
            # self.__start, self.__end = self.__end, self.__start # no longer reverse start and end points
            pass
        self.__direction = direction  # 1 or -1
        self._validate_direction()
        self._validate_region()
        self.start_extendable = False
        self.end_extendable = False

    @staticmethod
    def create_from_ends(left_end, right_end, context, direction=FORWARD, name=""):
        """Creates a region from the endpoints"""
        s, e = left_end, right_end
        if direction == Region.REVERSE:
            s, e = e, s
        return Region(s, e, context=context, direction=direction, name=name)

    def _validate_direction(self):
        """Validates that the direction key is understood"""
        if self.direction not in [Region.FORWARD, Region.REVERSE, Region.BOTH]:
            raise RegionError(
                "Direction {} not understood. Direction must be Region.FORWARD = {}, Region.REVERSE = {},\
             or Region.BOTH = {}".format(
                    self.direction, Region.FORWARD, Region.REVERSE, Region.BOTH
                )
            )

    def _validate_region(self):
        """Validates that the start and end regions are within the bounds of its context."""
        if (
            not self.context.circular
            and self.start > self.end
            and self.direction == Region.FORWARD
        ):
            raise RegionError("START cannot be greater than END for linear regions.")
        if (
            not self.context.circular
            and self.start < self.end
            and self.direction == Region.REVERSE
        ):
            raise RegionError("START cannot be greater than END for linear regions.")

    @property
    def start_index(self):
        """The minimum index for this region (defined by context). Alias of bounds_start"""
        return self.bounds_start

    @property
    def bounds_start(self):
        """The maximum index for the context of this region"""
        return self.context.start

    @property
    def bounds_end(self):
        """The maximum index for the context of this region"""
        return self.context.end

    @property
    def context_length(self):
        """The length of the context this region is in"""
        return self.context.length

    @property
    def length(self):
        """The length of this region"""
        if self._spans_origin():
            return self.context.span(
                self.rp, self.lp
            )  # if it spans origin, reverse span calculation
        else:
            return self.context.span(self.lp, self.rp)

    @property
    def circular(self):
        return self.context.circular

    @property
    def start(self):
        """
        Gets the start position of the region. Internally
        reverses the start and end positions if direction is
        reversed.

        :return:
        """
        return self.__start

    @start.setter
    def start(self, x):
        """
        Sets the start position of the region. Internally
        reverses the start and end positions if direction is
        reversed.

        :return:
        """
        self.__start = self.context.translate_pos(x)

    @property
    def end(self):
        """
        Gets the end position of the region. Internally
        reverses the start and end positions if direction is
        reversed.

        :return:
        """
        return self.__end

    @end.setter
    def end(self, x):
        """
        Sets the end position of the region. Internally
        reverses the start and end positions if direction is
        reversed.

        :return:
        """
        self.__end = self.context.translate_pos(x)

    @property
    def rp(self):
        """The right most point. The maximum index in the region."""
        return max(self.start, self.end)

    @property
    def lp(self):
        """The left most point. The minimum index in the region."""
        return min(self.start, self.end)

    @lp.setter
    def lp(self, v):
        """ Sets the left most point """
        if self.start < self.end:
            self.start = v
        elif self.end < self.start:
            self.end = v
        elif self.end == self.start:
            self.start = v
            self.end = v

    @rp.setter
    def rp(self, v):
        """ Sets the right most point """
        if self.start > self.end:
            self.start = v
        elif self.end > self.start:
            self.end = v
        elif self.end == self.start:
            self.start = v
            self.end = v

    @property
    def direction(self):
        """Direction of the region"""
        return self.__direction

    @property
    def left_end(self):
        """
        Returns left end stop

        e.g. ::

            |--------|
            ^L       ^R

            --|        |---
              ^R       ^L

        :return: index of left region end stop
        :rtype: int
        """
        if self._spans_origin():
            return self.rp
        else:
            return self.lp

    @property
    def right_end(self):
        """
        Returns left end stop
        e.g. ::

            |--------|
            ^L       ^R

            --|        |---
              ^R       ^L
        :return: index of right region end stop
        :rtype: int
        """
        if self._spans_origin():
            return self.lp
        else:
            return self.rp

    @left_end.setter
    def left_end(self, x):
        """Sets the left end stop"""
        if self.left_end == self.start:
            self.start = x
        elif self.left_end == self.end:
            self.end = x

    @right_end.setter
    def right_end(self, x):
        """Sets the right end stop"""
        if self.right_end == self.start:
            self.start = x
        elif self.right_end == self.end:
            self.end = x

    def is_forward(self):
        """Whether this region is pointed 'forward'"""
        d = self.direction
        return d in [Region.FORWARD]

    def is_reverse(self):
        """Whether this region is pointed in 'reverse'"""
        d = self.direction
        return d in [Region.REVERSE]

    def _spans_origin(self):
        """Returns whether region spans origin. Agnostic as to self.circular. """
        if self.start > self.end and self.direction == Region.FORWARD:
            return True
        if self.end > self.start and self.direction == Region.REVERSE:
            return True
        return False

    def _valid_indices(self, inclusive=True):
        """
        Returns tuples of valid indices within this region. Handles circular regions. ::

            E.g. |---------|
                 123456789
                   |>>>|
                   [(3,7)] if inclusive=True
                   [(4,6)] if inclusive=False

            E.g. |---------|
                 123456789
                   ||
                   [(3,4)] if inclusive=True
                   [] if inclusive=False

            E.g. |-------|
                 123456789
                 >>| |>>>>
                   [(1,3),(5,9)] if inclusive=True
                   [(1,2),(6,9)] if inclusive=False

            E.g. |-------|
                 123456789
                 |       |>
                [(1,1),(9,9)] if inclusive=True
                [] if inclusive=False

        :param inclusive: Whether region in inclusive
        :type inclusive: bool
        :return: list of tuples indicating valid index ranges for this region.
        :rtype: list
        """
        if not self._spans_origin():
            v = [(self.lp, self.rp)]
            if not inclusive:
                v = [(self.lp + 1, self.rp - 1)]
            if v[0][0] > v[0][1]:  # there are no valid indices
                v = []
            return v
        else:
            ranges = []
            left = (self.bounds_start, self.lp)
            right = (self.rp, self.context.end)
            if not inclusive:
                left = (self.context.start, self.lp - 1)
                right = (self.rp + 1, self.context.start)
            if left[0] <= left[1]:
                ranges.append(left)
            if right[0] <= right[1]:
                ranges.append(right)
            return ranges

    def within_region(self, pos, inclusive=True):
        """
        Returns whether a position is within the region. Handles circular regions.

        :param pos: index
        :type pos: int
        :param inclusive: whether to be inclusive (or exclusive)
        :type inclusive: bool
        :return: whether index is within region
        :rtype: bool
        """
        v = self._valid_indices(inclusive=inclusive)
        x = [i[0] <= pos <= i[1] for i in v]
        return any(x)

    def sub_region(self, s, e):
        """
        Creates a sub region with the same context_length and direction as this region.

        :param s: start index
        :type s: int
        :param e: end index
        :type e: int
        :return: creates a sub region of this region
        :rtype: Region
        """
        if self.within_region(s, inclusive=True) and self.within_region(
            e, inclusive=True
        ):
            r = self.copy()
            r.start = s
            r.end = e
            r._validate_region()
            return r
        else:
            raise RegionError(
                "Sub region bounds [{}-{}] outside of Region bounds [{}-{}]".format(
                    s, e, self.context.start, self.context.end
                )
            )

    def same_context(self, other):
        """
        Compares the context properties (length, bounds_start, bounds_ends) between two Regions.

        :param other: The other Region
        :type other: Region
        :return: Whether the other region's context sequence is the same length and has same start_index as this Region
        :rtype: bool
        """
        return self.context == other.context

    # TODO: Why not __copy__?
    def copy(self):
        """
        Creates another region with identical properties.

        :return: Copied region
        :rtype: Region
        """
        # s, e = self.start, self.end
        # if self.direction is Region.REVERSE:
        #     s, e = self.end, self.start
        return self.__class__(
            self.start, self.end, self.context, direction=self.direction, name=self.name
        )

    # TODO: Why does this copy itself??
    @force_same_context(error=True)
    def get_overlap(self, other):
        """
        Returns a region representing the overlap with another region
        e.g. ::

                |--------|         self
                      |--------|   other
                      |--| << This is returned

        :param other: other Region
        :type other: Region
        :return: Region if there is an overlap, None if there is no overlap
        :rtype: Region
        """
        if self.end_overlaps_with(other):
            r = self.copy()
            r.__direction = Region.FORWARD
            r.start = other.left_end
            r.end = self.right_end
            return r
        else:
            return None

    @force_same_context(error=True)
    def equivalent_location(self, other):
        return other.start == self.start and other.end == self.end

    @force_same_context(error=True)
    def encompasses(self, other, inclusive=True):
        return self.within_region(
            other.start, inclusive=inclusive
        ) and self.within_region(other.end, inclusive=inclusive)

    @force_same_context(error=True)
    def end_overlaps_with(self, other):
        """
         Whether this region overlaps the next region it this regions end. Other needs some kind of overhang to
         return True.
         False if context is different. ::

             True
                 self   |------|
                 other      |-------|

             False
                 self         |------|
                 other  |-------|

             False
                 self   |------|
                 other    |----|

             True
                 self   |------|
                 other    |-----|
        :param other: other Region
        :type other: Region
        :return: True if region ends overlaps, False if otherwise.
        :rtype: bool
        """

        return self.within_region(
            other.left_end, inclusive=True
        ) and not self.encompasses(other)

    @force_same_context(error=True)
    def get_gap(self, other):
        """
        Gets the gap region. Returns None if there is no gap or Regions are not consecutive. Always are in the
        forward direction. ::

            Context:        |----------------------|
            This Region:        |-------|
            Other Region:                    |------|
            Gap:                         |==|


            r1:         ----|              |-----
            r2:                  |----|
            r1.get_gap(r2)   |==|
            r2.get_gap(r1)             |==|

        :param other: other Region
        :type other: Region
        :return: gap as a Region
        :rtype: Region
        """
        if self.consecutive_with(other):
            return None
        if not self.within_region(other.lp) or self is other:
            r = self.copy()
            try:
                r.end = other.left_end - 1
                r.start = self.right_end + 1
            except RegionError:
                return None
            return r
        else:
            return None

    @force_same_context(error=True)
    def get_gap_span(self, other):
        """
        Returns span of gap. Returns 0 if regions are consecutive, negative if regions overlap, positive for gaps.

        e.g. ::

            +        |--------|***|-----| (length of three gap)

            -        |------***
                            ***-----|
                            *** (length of three overhang)

            0        |----||----| (0 for consecutive)

            None     |--------|
                        |-----| (None

        :param other:
        :type other:
        :return: span of gap (0 for consecutive, - for overlap, + for gap)
        :rtype: int
        """
        overlap = self.get_overlap(other)
        gap = self.get_gap(other)
        cons = self.consecutive_with(other)
        if cons:
            return 0
        if overlap is not None:
            return -overlap.length
        if gap is not None:
            return gap.length

    # @force_same_context(error=True)
    # def no_overlap(self, other):
    #     """
    #     Returns True if there is no overlap with the other Region
    #
    #     :param other: Other Region
    #     :type other: Region
    #     :return: if there is no overlap
    #     :rtype: bool
    #     """
    #     return not self.within_region(other.start, inclusive=True) \
    #            and not other.within_region(self.end, inclusive=True)

    @force_same_context(error=True)
    def consecutive_with(self, other, ignore_direction=True):
        """
        Returns whether the right_end is consecutive with the other region's left_end ::

            |-------||----|

        :param other: other Region
        :type other: Region
        :return: True if other region is consecutive with this region
        :rtype: bool
        """

        expected_right_end = None
        expected_left_end = None
        try:
            expected_left_end = self.context.translate_pos(self.right_end + 1)
        except RegionError:
            return False
        try:
            expected_right_end = other.context.translate_pos(other.left_end - 1)
        except RegionError:
            return False
        consecutive = (
            self.right_end == expected_right_end and other.left_end == expected_left_end
        )
        return consecutive

    @force_same_context(error=True)
    def fuse(self, other, inplace=True):
        """
        Fuses this region with other region. If regions are not consecutive, raises RegionError ::

            |-----------| (1)
                         |-----------| (2)
            |------------------------| (1 modified)

        :param other: other Region
        :type other: Region
        :return: this Region (self, modified by extension by other Region)
        :rtype: Region
        """
        if self.consecutive_with(other):
            new_region = self
            if not inplace:
                new_region = self.copy()
            new_region.right_end = other.right_end
            return new_region
        else:
            return None
            # raise RegionError(
            #         "Cannot fuse regions [{}-{}] with [{}-{}].".format(self.start, self.end, other.start, other.end))

    def extend_start(self, x):
        """
        Extends the start by x amount. Retracts start if negative. Raises error if retracts past end.

        e.g. extend start by +4 ::

                s          e
                |----------|
            |<<<|----------|
            |--------------|

        :param x:
        :type x:
        :return:
        :rtype:
        """
        self._extend(x, self.start)

    def extend_end(self, x):
        """
        Extends the end by x amount. Retracts end if negative. Raises error if retracts past start.

        e.g. extend end by +4 ::

            s          e
            |----------|
            |----------|>>>|
            |--------------|

        :param x:
        :type x:
        :return:
        :rtype:
        """
        self._extend(x, self.end)

    def extend_left_end(self, x):
        """ Extends (or retracts) the left end by x amount. """
        self._extend(x, self.left_end)

    def extend_right_end(self, x):
        """ Extends (or retracts) the right end by x amount. """
        self._extend(x, self.right_end)

    def _extend(self, x, end_pos):
        """ Extends or retracts by x amount at end_pos. """
        if x < -self.length + 1:
            raise RegionError(
                "Cannot retract end past region start (start: {0}, end: {1}, x: {2})".format(
                    self.start, self.end, x
                )
            )
        if x > self.context.length - self.length:
            raise RegionError(
                "Cannot extend end around origin and past other end (start: {0}, end: {1}, "
                "x: {2})".format(self.start, self.end, x)
            )
        if end_pos == self.right_end:
            self.right_end += x
        elif end_pos == self.left_end:
            self.left_end -= x
        else:
            raise RegionError(
                "Position at {0} not at either end. Cannot extend.".format(end_pos)
            )

    def set_forward(self):
        """ Reverses direction of region if region is reverse """
        if not self.is_forward():
            self.reverse_direction()

    def set_reverse(self):
        """ Reverses direction of region if region is forward """
        if self.is_forward():
            self.reverse_direction()

    def reverse_direction(self):
        """ Reverses the direction of this region. Reverses the start and end indices. """
        self.end, self.start = self.start, self.end
        if self.is_forward():
            self.__direction = Region.REVERSE
        else:
            self.__direction = Region.FORWARD
        return self.direction

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __len__(self):
        return self.length

    def __str__(self):
        return "Region(start={start} end={end}, name={name}, direction={direction}, context={context})".format(
            start=self.start,
            end=self.end,
            direction=self.direction,
            context=self.context,
            name=self.name,
        )

    def __repr__(self):
        return str(self)



from abc import ABC, abstractmethod


class ClassInterface(ABC):
    def __init__(self, cls):
        self._cls = cls

    @abstractmethod
    def instance(self, instance):
        pass


class Interface(object):
    def __init__(self, instance, class_interface):
        self._inst = instance
        self._class_interface = class_interface



import functools
from enum import Flag
import itertools
import operator


class NumberStrategy(Flag):
    NUM = "line_number"
    LENGTH = "line_length"


def chunks(iterable, n):
    chunk = None
    for i, x in enumerate(iterable):
        if i % n == 0:
            if chunk is not None:
                yield chunk
            chunk = []
        chunk.append(x)
    yield chunk


def to_lines(string, width):
    """Converts a string to lines of length <= width"""
    lines = []
    for i in range(0, len(string), width):
        lines.append(string[i : i + width])
    return lines


def rows_to_chunks(rows, width=75):
    if not rows:
        return []
    lengths = set([len(r) for r in rows])
    if len(lengths) > 1:
        raise Exception("Cannot format rows that have different lengths")

    lines = []
    for row in rows:
        line = to_lines(row, width=width)
        lines.append(line)
    #     print(lines)
    interleafed = functools.reduce(lambda x, y: x + y, zip(*lines))
    return list(chunks(interleafed, len(rows)))


def prepend_lines(lines, label_iterable, indent, fill=" ", align="<"):
    """

    :param lines:
    :type lines:
    :param indent:
    :type indent: int
    :param fill: default ' '
    :type fill: what to fill the spaces
    :param align: either left "<", center "^" or right ">"
    :type align: string
    :return:
    :rtype:
    """
    prepend_pattern = functools.partial(
        "{0:{fill}{align}{indent}}".format, fill=fill, align=align, indent=indent
    )
    new_lines = []
    for label, line in zip(label_iterable, lines):
        print(line)
        new_lines.append("{}{}".format(prepend_pattern(label), line))
    return new_lines


def indent(lines, indent):
    return prepend_lines(lines, [""] * len(lines), indent)


def set_indent(lines, indent):
    return indent([l.lstrip() for l in lines], indent)


def enumerate_lines(lines, indent):
    labels = range(len(lines))
    return prepend_lines(lines, labels, indent)


def accumulate_length_of_lines(lines, indent):
    labels = itertools.accumulate([len(l.strip("\n")) for l in lines], operator.add)
    return prepend_lines(lines, labels, indent)


def accumulate_length_of_first_line(lines, indent):
    labels = itertools.accumulate(
        [len(l.split("\n")[0].strip("\n")) for l in lines], operator.add
    )
    return prepend_lines(lines, labels, indent)


def number_lines(lines, indent=None, numbering_strategy=NumberStrategy.LENGTH, step=1):
    if indent is None:
        indent = 10
    index = 0
    indices = []
    for i, l in enumerate(lines):
        if i % step == 0:
            indices.append(index)
            if numbering_strategy == NumberStrategy.LENGTH:
                index += len(l)
            elif numbering_strategy == NumberStrategy.NUM:
                index += 1
        else:
            indices.append("")

    new_lines = []
    for index, line in zip(indices, lines):
        index_str = "{0:{fill}{align}{indent}}".format(
            index, fill=" ", align="<", indent=indent
        )
        new_lines.append("{}{}".format(index_str, line))
    return new_lines


def interleaf(rows, width=75, spacer=None, number=False, indent=None):
    """Interleaf lines that have the same lengths"""
    if not rows:
        return ""
    lengths = set([len(r) for r in rows])
    if len(lengths) > 1:
        raise Exception("Cannot format rows that have different lengths")

    lines = []
    for row in rows:
        line = to_lines(row, width=width)
        lines.append(line)
    interleafed = functools.reduce(lambda x, y: x + y, zip(*lines))

    if number:
        interleafed = number_lines(interleafed, step=len(rows), indent=indent)

    with_spacer = []
    for i, l in enumerate(interleafed):
        if spacer is not None and i % len(rows) == 0:
            with_spacer.append(spacer)
        with_spacer.append(l)

    return "\n".join(with_spacer)


def group_by_line_lengths(lines):
    # group by line len
    groups = [[]]
    l = len(lines[0])
    for line in lines:
        if len(line) != l:
            groups.append([])
        if line:
            groups[-1].append(line)
    return groups


def format_sequence(s, width=75, spacer=""):
    groups = group_by_line_lengths(s.split("\n"))
    return "\n".join(
        [interleaf(g, width=width, spacer=spacer, number=True) for g in groups]
    )



"""
Methods for converting jdna objects to other formats
"""
from Bio import SeqIO
from jdna.viewer import FASTAViewer
from jdna.interface import ClassInterface, Interface


class IOInterface(ClassInterface):
    def from_seqrecord(self, seqrecord):
        return self._cls(
            str(seqrecord.seq), name=seqrecord.id, description=seqrecord.description
        )

    @staticmethod
    def fasta(sequences, out=None):
        fasta_viewer = FASTAViewer(sequences)
        if out:
            with open(out, "w") as f:
                f.write(str(fasta_viewer))
        return fasta_viewer

    @classmethod
    def write_fasta(cls, sequences, out):
        return cls.fasta(sequences, out)

    def parse(self, inpath, format):
        seq_gen = SeqIO.parse(inpath, format=format)
        return [self.from_seqrecord(seq) for seq in seq_gen]

    def read(self, inpath, format):
        seq = SeqIO.read(inpath, format=format)
        return self.from_seqrecord(seq)

    def read_fasta(self, inpath):
        return self.parse(inpath, format="fasta")

    def instance(self, instance):
        return IOInstanceInterface(instance, self)


class IOInstanceInterface(Interface):
    def fasta(self, *sequences, out=None):
        return self._class_interface.fasta([self._inst] + sequences, out=out)

    def write_fasta(self, out):
        return self.fasta(out)

    def json(self, *args, **kwargs):
        return self._inst.json(*args, **kwargs)



"""
class representing base pairs and their complements
"""

import random
from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
import re


class Alphabet(object):
    """
    A dictionary class that retrieves complementary base pairs
    """

    __slots__ = ["_chr", "_comp", "__complementary", "__ambiguous"]

    def __init__(self, characters, complementary_characters, ambiguous_characters=None):
        self.__complementary = dict(
            zip(
                characters.lower() + characters.upper(),
                complementary_characters.lower() + complementary_characters.upper(),
            )
        )
        if ambiguous_characters is None:
            ambiguous_characters = {}
        self.__ambiguous = ambiguous_characters

    @classmethod
    def from_biopython_alphabet(
        cls, biopython_alphabet: IUPAC.Alphabet, ambiguous_characters=None
    ):
        seq = Seq(biopython_alphabet.letters)
        seq_letters = str(seq)
        c_seq_letters = str(seq.complement())
        return cls(seq_letters, c_seq_letters, ambiguous_characters)

    def characters(self):
        return self.__complementary.keys()

    def random(self):
        """
        Return random character
        """
        return random.choice(list(self.__complementary.keys()))

    def compare(self, s1, s2, ignore_case=True):
        """
        Compare two sequences. If the second sequence has ambiguous bases, convert
        second sequence to a regex to compare with the first sequence.

        :param s1:
        :type s1:
        :param s2:
        :type s2:
        :return:
        :rtype:
        """
        pattern = "".join(self.__ambiguous.get(b, b) for b in s2)
        if ignore_case:
            match = re.match(pattern, s1, re.IGNORECASE)
        else:
            match = re.match(pattern, s1)
        if match:
            return True
        return False

    def complement(self, basestring):
        return "".join(self[x] for x in basestring)

    def reverse_complement(self, basestring):
        return self.complement(basestring)[::-1]

    def __getitem__(self, item):
        return self.__complementary[item]

    def __contains__(self, item):
        return item in self.__complementary

    # aliases
    rc = reverse_complement
    c = complement


_iupacdict = {
    "M": "[AC]",
    "R": "[AG]",
    "W": "[AT]",
    "S": "[CG]",
    "Y": "[CT]",
    "K": "[GT]",
    "V": "[ACG]",
    "H": "[ACT]",
    "D": "[AGT]",
    "B": "[CGT]",
    "X": "[ACGT]",
    "N": "[ACGT]",
}

iupacdict = {}
for k, v in _iupacdict.items():
    iupacdict[k.lower()] = v.lower()
    iupacdict[k.upper()] = v.upper()

AmbiguousDNA = Alphabet.from_biopython_alphabet(IUPAC.ambiguous_dna, iupacdict)
UnambiguousDNA = Alphabet.from_biopython_alphabet(IUPAC.unambiguous_dna, {})



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['jdna', 'jdna._version', 'jdna.regions']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.73,<2.0',
 'colorama>=0.4.1,<0.5.0',
 'networkx>=2.3,<3.0',
 'numpy>=1.16,<2.0',
 'primer3-py>=0.6.0,<0.7.0',
 'webcolors>=1.9,<2.0']

entry_points = \
{'console_scripts': ['name = jdna:_version.get_name',
                     'upver = jdna:_version.pull_version',
                     'version = jdna:_version.get_version']}

setup_kwargs = {
    'name': 'jdna',
    'version': '0.1.2',
    'description': '',
    'long_description': '# jdna\n\nA Python DNA sequence editor that represents sequences as a linked list.\nIt includes basic molecular reaction simulations (primer annealing, pcr, dna assembly), basic format conversion, and a viewer class.\n\n[Documentation is here](https://jvrana.github.io/jdna/index)\n\n![example.png](./docsrc/_static/viewer.png)\n',
    'author': 'Justin Vrana',
    'author_email': 'justin.vrana@gmail.com',
    'url': 'https://github.com/jvrana/jdna',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""dna sequence editor

==========
jdna
==========

.. autosummary::
    :toctree: _autosummary

    linked_list
    sequence
    alphabet
    viewer
    reaction
    utils

"""

from ._version import __version__, __title__, __author__, __homepage__, __repo__
from jdna.linked_list import Node, DoubleLinkedList
from jdna.alphabet import UnambiguousDNA, AmbiguousDNA
from jdna.sequence import Feature, Sequence, Nucleotide
from jdna.reaction import Reaction
from jdna.viewer import SequenceViewer



"""
misc utilities
"""

import random

import webcolors
from colorama import Fore, Style, Back

colors_rgb = {}
for c in vars(Fore):
    try:
        colors_rgb[c] = webcolors.name_to_rgb(c)
    except:
        pass
colors_rgb


def rgb_to_hex(r, g, b):
    def clamp(x):
        return max(0, min(x, 255))

    return "#{0:02x}{1:02x}{2:02x}".format(clamp(r), clamp(g), clamp(b))


def hex_to_color_name(h):
    """Get closest named color for terminal printing"""
    distance = 3 * 255 ** 2 + 1
    color = None
    for name, rgb in colors_rgb.items():
        rgb2 = webcolors.hex_to_rgb(h)
        d = 0
        for _c in ["red", "green", "blue"]:
            d += (getattr(rgb2, _c) - getattr(rgb, _c)) ** 2
        if d < distance:
            distance = d
            color = name
    return color


def _colored(text, hex_color, fore_or_back):
    """Return colored text"""
    if hex_color is None:
        return text
    try:
        color_name = hex_to_color_name(hex_color)
    except ValueError:
        color_name = hex_color
    s = "{color}{text}{end}".format(
        color=getattr(fore_or_back, color_name.upper()), text=text, end=Style.RESET_ALL
    )
    return s


def colored(text, hex_color):
    """Return colored text"""
    return _colored(text, hex_color, Fore)


def colored_background(text, hex_color):
    """Return colored text"""
    return _colored(text, hex_color, Back)


def random_color():
    rgb = [int(random.random() * 255) for x in range(3)]
    return rgb_to_hex(*rgb)



"""
Classes to view sequences.

The viewer can display sequences and annotations, as in the following:

.. code::


    > "Unnamed" (550bp)


                                                                ----------------GFP----------------
                                                                |<START
                                                                ----      -----------RFP-----------
    0         CCCAGGACTAGCGACTTTCCGTAACGCGACCTAACACCGGCCGTTCCTTCGAGCCAGGCAAATGTTACGTCACTTCCTTAGATTT
              GGGTCCTGATCGCTGAAAGGCATTGCGCTGGATTGTGGCCGGCAAGGAAGCTCGGTCCGTTTACAATGCAGTGAAGGAATCTAAA

              ------GFP------
              -----------------------------------------RFP-----------------------------------------
    85        TGAACAGCGCCGTACCCCGATATGATATTTAGATATATAGCAGTTACACTTGGGGTTGCTATGGACTTAGATCTGCTGTATGTTT
              ACTTGTCGCGGCATGGGGCTATACTATAAATCTATATATCGTCAATGTGAACCCCAACGATACCTGAATCTAGACGACATACAAA

              -----------------------------------------RFP-----------------------------------------
    170       TCTTACCTTCCGCATCAGGGGACAATTCGCCAGTAGAATTCAGTTTGTGCGTGAGAACATAAGATTGAATCCCACGCAGGCACAA
              AGAATGGAAGGCGTAGTCCCCTGTTAAGCGGTCATCTTAAGTCAAACACGCACTCTTGTATTCTAACTTAGGGTGCGTCCGTGTT

              ---------------------RFP----------------------
    255       GCAGGGCGGGCAGACTCTATAGGTCCTAAGACCCTGAGACTGCGTCCTCAAGATACAGGTTAACAATCCCCGTATGGAGCCGTTC
              CGTCCCGCCCGTCTGAGATATCCAGGATTCTGGGACTCTGACGCAGGAGTTCTATGTCCAATTGTTAGGGGCATACCTCGGCAAG

    340       TTAGCATGACCCGACAGGTGGGCTTGGCTCGCGTAAGTTGAGTGTTGCAGATACCTGCTGCTGCGCGGTCTAGGGGGAATCGCCG
              AATCGTACTGGGCTGTCCACCCGAACCGAGCGCATTCAACTCACAACGTCTATGGACGACGACGCGCCAGATCCCCCTTAGCGGC

    425       ATTTTGACGTAGGATCGGTAATGGGCAGTAAACCCGCAACTATTTTCAGCACCAGATGCAAGTTTCCCTAGAAAGCGTCATGGTT
              TAAAACTGCATCCTAGCCATTACCCGTCATTTGGGCGTTGATAAAAGTCGTGGTCTACGTTCAAAGGGATCTTTCGCAGTACCAA

    510       TGCAATCTCCTTAGGTCACAGCAAACATAGCAGCCCCTGT
              ACGTTAGAGGAATCCAGTGTCGTTTGTATCGTCGGGGACA
"""

import re
import functools
import itertools
from collections import OrderedDict

from networkx import nx

from jdna.utils import random_color, colored, colored_background


class StringColumn(object):
    """Class for managing string columns"""

    FILL = " "

    def __init__(self, strings=None, color=None, background=None, fill=None):
        """
        StringColumn constructor.

        :param strings: list of strings
        :type strings: list
        """
        self._strings = []
        self._length = 0
        if fill is None:
            fill = self.FILL
        self.fill = fill
        if strings:
            max_length = max([self.string_length(s) for s in strings])
            self._length = max_length
            for s in strings:
                self.append_string(self.right_fill(s))
        self.color = color
        self.background = background

    def apply_color(self, color):
        self._strings = [colored(s, color) for s in self._strings]
        return self

    def apply_background_color(self, color):
        self._strings = [colored_background(s, color) for s in self._strings]
        return self

    @property
    def length(self):
        return self._length

    @staticmethod
    def remove_formatting(string):
        pattern = "\\x1b\[\d\dm"
        return re.sub(pattern, "", string)

    @classmethod
    def string_length(cls, string):
        """String length, ignoring terminal formatting"""
        subbed = cls.remove_formatting(string)
        l = len(subbed)
        return len(subbed)

    @property
    def strings(self):
        return self._strings[:]

    def indent(self, num):
        sc_copy = self.copy()
        for i, s in enumerate(sc_copy.strings):
            sc_copy._strings[i] = self.fill * num + sc_copy._strings[i]
        sc_copy._length += num
        return sc_copy

    def indent_right(self, num):
        sc_copy = self.copy()
        for i, s in enumerate(sc_copy.strings):
            sc_copy._strings[i] = sc_copy._strings[i] + self.fill * num
        sc_copy._length += num
        return sc_copy

    def center(self, span):
        diff = span - self.length
        if diff > 0:
            l = int(diff / 2)
            r = l + diff % 2
            return self.indent(l).indent_right(r)
        return self.copy()

    def flip(self):
        self._strings = self._strings[::-1]
        return self.copy()

    def right_fill(self, string):
        return string + self.fill * (self.length - self.string_length(string))

    def prepend_string(self, new_string):
        if self.string_length(new_string) > self.length:
            self._length = self.string_length(new_string)
        self._strings.insert(0, self.right_fill(new_string))

    def append_string(self, new_string):
        if self.string_length(new_string) > self.length:
            self._length = self.string_length(new_string)
        self._strings.append(self.right_fill(new_string))

    def add_prefix(self, prefix):
        for i, s in self.strings:
            self.strings[i] = prefix + self.strings[i]

    def stack(self, *others):
        sc = self.copy()
        for other in others:
            for string in other.strings:
                sc.append_string(string)
        return sc

    def __contains__(self, item):
        return any([item in s for s in self.strings])

    def __add__(self, other):
        if isinstance(other, str):
            other = StringColumn([other])
        else:
            other = other.copy()
        sc = self.copy()

        diff = len(sc.strings) - len(other.strings)
        if diff > 0:
            for i in range(diff):
                other.prepend_string("")
        elif diff < 0:
            for i in range(-diff):
                sc.prepend_string("")

        new_sc = self.copy_empty()
        for this_string, other_string in zip(sc.strings, other.strings):
            new_sc.append_string(this_string + other_string)
        return new_sc

    def copy_empty(self):
        sc_copy = self.copy()
        sc_copy._strings = []
        return sc_copy

    def copy(self):
        return self.__copy__()

    def strip_indices(self):
        n1 = 0
        n2 = 0
        for x in self[:]:
            if all([_x == " " for _x in x]):
                n1 += 1
            else:
                break
        for x in self[::-1]:
            if all([_x == " " for _x in x]):
                n2 += 1
            else:
                break
        return n1, len(self) - n2

    def strip(self):
        n1, n2 = self.strip_indices()
        return self[n1:n2]

    def __copy__(self):
        copied = self.__class__(
            self.strings, color=self.color, background=self.background, fill=self.fill
        )
        return copied

    def __getitem__(self, key):
        # strings = [self.remove_formatting(s) for s in self.strings]
        strings = [s.__getitem__(key) for s in self.strings]
        string_col = self.__class__(strings)
        return string_col

    #     def __setitem__(self, key, items):
    #         if not len(items) == len(self.strings):
    #             raise TypeError("Value must have {} items".format(len(self.strings)))
    #         for string, item in zip(self.strings, items):
    #             string[key] = item

    def __eq__(self, other):
        return str(self) == str(other)

    def __iter__(self):
        return zip(*self.strings)

    def __len__(self):
        return self.length

    def __str__(self):
        s = "\n".join(self.strings)
        return s

    def __repr__(self):
        return str(self)

    @classmethod
    def condense(cls, rows):
        """
        Condense a list of :class:`StringColumn` into the minimum number of StringColumns comprising of columns stripped
        of white space. Briefly, this is similar to the following procedure:

        .. code-block::

            input = [
                'label         ',
                '       label2 ',
                '      label3  '
            ]

            # >> CONDENSE

            output = [
                'label  label2 ',
                '      label3  '
            ]

        :param rows:
        :type rows:
        :return:
        :rtype:
        """
        segments = []
        indexed_segments = []
        previous_end = 0
        for row in rows:
            col = (
                row.strip()
                .apply_color(row.color)
                .apply_background_color(row.background)
            )
            start, end = row.strip_indices()
            word = (start, end, col)
            if word not in segments:
                segments.append(tuple(list(word)))
                indexed_segments.append((start, end, col, previous_end))
                previous_end += 1

        # create a graph of non-overlapping segments
        nonoverlap_graph = nx.Graph()
        for w in indexed_segments:
            nonoverlap_graph.add_node(w[-1])
        for segment1, segment2 in itertools.combinations(indexed_segments, 2):
            start1, end1, _, index1 = segment1
            start2, end2, _, index2 = segment2
            if start1 < start2 or start1 > end2:
                if start2 < start1 or start2 > end1:
                    if end1 < start2:
                        nonoverlap_graph.add_edge(index1, index2)
                    else:
                        nonoverlap_graph.add_edge(index2, index1)

        # find minimum number of cliques that covers the graph (clique covering)
        subgraph = nonoverlap_graph.subgraph(nonoverlap_graph.nodes)
        cliques = []
        while len(subgraph):
            max_clique = list(nx.find_cliques(subgraph))[0]
            cliques.append(max_clique)
            remaining = set(subgraph.nodes).difference(set(max_clique))
            subgraph = nonoverlap_graph.subgraph(list(remaining))

        condensed_rows = []
        for clique in cliques:
            string_column = cls()
            clique_segments = [indexed_segments[s] for s in clique]
            clique_segments = sorted(clique_segments, key=lambda seg: seg[1])

            previous_end = 0
            for segment in clique_segments:
                start, end, seg_str_col, _ = segment
                string_column += seg_str_col.indent(start - previous_end)
                previous_end = end
            condensed_rows.append(string_column)
        return condensed_rows


def chunkify(iterable, n):
    """Break an interable into chunks of size at most 'n'"""
    chunk = None
    for i, x in enumerate(iterable):
        if i % n == 0:
            if chunk is not None:
                yield chunk
            chunk = []
        chunk.append(x)
    yield chunk


def to_lines(string, width):
    """Converts a string to lines of length <= width"""
    lines = []
    for i in range(0, len(string), width):
        lines.append(string[i : i + width])
    return lines


def prepend_lines(lines, label_iterable, indent, fill=" ", align="<"):
    """
    Prepend lines with a label

    :param lines: lines to prepend
    :type lines: list
    :param indent: number of spaces between start of label and start of line
    :type indent: int
    :param fill: default ' '
    :type fill: what to fill the spaces
    :param align: either left "<", center "^" or right ">"
    :type align: string
    :return: new prepended lines
    :rtype: list
    """
    prepend_pattern = functools.partial(
        "{0:{fill}{align}{indent}}".format, fill=fill, align=align, indent=indent
    )
    new_lines = []
    for label, line in zip(label_iterable, lines):
        new_lines.append("{}{}".format(prepend_pattern(label), line))
    return new_lines


def indent(string, indent):
    """Indent lines"""
    lines = string.split("\n")
    new_lines = prepend_lines(lines, [""] * len(lines), indent)
    return "\n".join(new_lines)


# def set_indent(lines, indent):
#     """Reset the indent of lines"""
#     return indent([l.lstrip() for l in lines], indent)
#
#
# def enumerate_lines(lines, indent):
#     """Enumerate lines"""
#     labels = range(len(lines))
#     return prepend_lines(lines, labels, indent)
#
#
# def accumulate_length_of_lines(lines, indent):
#     labels = itertools.accumulate([len(l.strip('\n')) for l in lines], operator.add)
#     return prepend_lines(lines, labels, indent)
#
#
# def accumulate_length_of_first_line(lines, indent):
#     labels = itertools.accumulate([len(l.split('\n')[0].strip('\n')) for l in lines], operator.add)
#     return prepend_lines(lines, labels, indent)


class ViewerAnnotationFlag(object):
    """Flags for annotation directions"""

    FORWARD = ">"
    REVERSE = "<"
    BOTH = "-"


class SequenceRow(object):
    """A row in a :class:`SequenceViewer` instance. Can be comprised of multiple sequences (i.e. lines)
    and can be annotated with 'features'."""

    def __init__(
        self, lines, labels, indent, start, end, line_colors=None, line_backgrounds=None
    ):
        """
        SequenceRow constructor

        :param lines: list of lines to display. Lengths of all lines must all be equivalent.
        :type lines: list
        :param labels: list of labels to apply to each line
        :type labels: list
        :param indent: indent to apply to the lines
        :type indent: string
        :param start: start bp of this row
        :type start: int
        :param end: end bp of this row
        :type end: int
        """
        lengths = set([len(r) for r in lines])
        if len(lengths) > 1:
            raise Exception("Cannot format rows that have different lengths")
        self._lines = lines
        if isinstance(line_colors, str):
            line_colors = [line_colors] * len(lines)
        if isinstance(line_backgrounds, str):
            line_backgrounds = [line_backgrounds] * len(lines)
        self.line_colors = line_colors
        self.line_backgrounds = line_backgrounds
        self.labels = labels
        self.indent = indent
        self.start = start
        self.end = end
        self.annotations = []
        self.bottom_annotations = []

    @property
    def lines(self):
        lines = self._lines[:]
        if self.line_colors:
            lines = [
                colored(line, color) for line, color in zip(lines, self.line_colors)
            ]
        if self.line_backgrounds:
            lines = [
                colored_background(line, color)
                for line, color in zip(lines, self.line_backgrounds)
            ]
        return prepend_lines(lines, self.labels, self.indent)

    def annotation_lines(self, annotations):
        condensed = StringColumn.condense(annotations)
        return [str(a.indent(self.indent)) for a in condensed]

    @staticmethod
    def make_annotation(label, span, fill="*", color=None, background=None):
        """
        Make an annotation with 'label' spanning inclusive base pairs indices 'span'

        :param label: annotation label
        :type label: basestring
        :param span: the start and end (inclusive) of the annotation
        :type span: tuple
        :param fill: what to fill whitespace with
        :type fill: basestring
        :return:
        :rtype:
        """

        if len(fill) != 1:
            raise Exception(
                "Fill '{}' must be a single character long, not {} characters".format(
                    fill, len(fill)
                )
            )
        if fill.strip() == "":
            raise Exception("Fill cannot be whitespace")
        sc = StringColumn(color=color, background=background)
        if isinstance(label, str):
            if len(label) + 1 > span:
                sc.append_string(label)
                # sc.append_string("|<{0:{fill}{align}{indent}}".format(label, fill=' ', align='^', indent=span))
                label = fill * span
            sc2 = StringColumn(fill=fill)
            sc2.append_string(label)
            return sc.stack(sc2.center(span))
            # sc2.append_string("{0:{fill}{align}{indent}}".format(label, fill=fill, align='^', indent=span))
        elif isinstance(label, StringColumn):
            if len(label) > span:
                sc = sc.stack(label)
                return sc.stack(StringColumn([""], fill=fill).center(span))
            label.fill = fill
            return sc.stack(label.center(span))

    def absolute_annotate(
        self, start, end, fill, label, color=None, background=None, top=True
    ):
        """
        Applyt annotation to this row using absolute start and ends for
        THIS row.

        :param start: inclusive start
        :type start: int
        :param end: inclusive end
        :type end: int
        :param fill: what to fill whitespace with
        :type fill: basestring
        :param label: annotation label
        :type label: basestring
        :return: None
        :rtype: None
        """
        span = end - start + 1
        annotation = self.make_annotation(
            label, span, fill, color=color, background=background
        ).indent(start)
        if top:
            self.annotations.append(annotation)
        else:
            self.bottom_annotations.append(annotation)

    def annotate(
        self,
        start,
        end,
        fill,
        label="",
        color=None,
        background=None,
        top=True,
        wrap=False,
    ):
        """
        Annotate the sequence row. If 'start' or 'end' is beyond,
        the expected start or end for this row, the annotation will
        automatically be truncated.

        :param start: inclusive start
        :type start: int
        :param end: inclusive end
        :type end: int
        :param fill: what to fill whitespace with
        :type fill:
        :param label: optional label to apply to the annotation
        :type label: basestring
        :return:
        :rtype:
        """
        s = max(start - self.start, 0)
        e = min(end - self.start, len(self) - 1)
        return self.absolute_annotate(
            s, e, fill, label, color=color, background=background, top=top
        )

    def in_bounds(self, x):
        """
        Checks if the index 'x' is in between row start and end (inclusive)

        :param x: index
        :type x: int
        :return: if in bounds
        :rtype: bool
        """
        return x >= self.start and x <= self.end

    def __len__(self):
        return len(self._lines[0])

    def __str__(self):
        return "\n".join(
            self.annotation_lines(self.annotations)
            + self.lines
            + self.annotation_lines(self.bottom_annotations)
        )


#
# class SequenceLabel(object):
#
#     def __init__(self, indent, label=None, pattern=None, indexer=None):
#         self.indent = indent
#         self.index = 0
#         self.label = label
#         if pattern is None:
#             pattern = "{label} {index}"
#         self.pattern = pattern
#         self.indexer = indexer
#
#     def indexers(self):
#         return {
#             "line_length": lambda x: self.index + len(x),
#             "enumerate": lambda x: x + 1
#         }
#
#     def enumerate(self, line):
#         if self.indexer:
#             self.index += self.indexer(line)
#
#     def __str__(self):
#         label = self.patter.format(index=self.index, label=self.label)
#         return "{0:{fill}{align}{indent}".format(label, fill=' ', align='<', indent=self.indent)


class SequenceViewer(object):
    """A class that views longs sets of sequences."""

    class DEFAULTS:
        METADATA_INDENT = 2
        INDENT = 10
        SPACER = "\n"
        HEADER_SPACER = "\n"
        WIDTH = 85
        NAME = "Unnamed"
        DESCRIPTION = ""
        BACKGROUND_COLOR = None
        FOREGROUND_COLOR = None
        APPLY_INDICES = [0]

    RANDOM_COLOR = "RANDOM"

    def __init__(
        self,
        sequences,
        sequence_labels=None,
        apply_indices=DEFAULTS.APPLY_INDICES,
        foreground_colors=DEFAULTS.FOREGROUND_COLOR,
        background_colors=DEFAULTS.BACKGROUND_COLOR,
        indent=DEFAULTS.INDENT,
        width=DEFAULTS.WIDTH,
        spacer=DEFAULTS.SPACER,
        header_spacer=DEFAULTS.HEADER_SPACER,
        name=DEFAULTS.NAME,
        window=(0, None),
        description="",
        metadata=None,
    ):
        """
        SequenceViewer constructor

        :param sequences: list of sequences to view
        :type sequences: list
        :param sequence_labels: optional labels to apply to sequence. Include the '{index}' to enumerate the base pairs.
        :type sequence_labels: list
        :param foreground_colors: optional list base pair foreground colors (hex or common name) to apply to each sequence. If a string
                                    is provided, color will be applied to all sequences. If provided with "RANDOM",
                                    a random color will be assigned to each sequence.
        :type foreground_colors: list
        :param background_colors: optional list base pair background colors (hex or common name) to apply to each sequence.
                                    Usage is analogous to `foreground_colors` parameter.
        :type background_colors: list
        :param indent: spacing before start of string and start of base pairs
        :type indent: int
        :param width: width of the view window for the sequences (e.g. width=100 would mean rows of at most len 100
                        characters
        :type width: string
        :param spacer: string to apply inbetween rows (default is newline)
        :type spacer: string
        :param name: optional name for this viewer, to be displayed in the header
        :type name: basestring
        :param window: tuple of the start and end points of the viewing window
        :type window: tuple
        :param description: optional description for this viewer
        :type description: basestring
        :param metadata: optional metadata to display in the header
        :type metadata: dict
        """
        assert isinstance(sequences, list)
        seq_lens = set([len(s) for s in sequences])
        if len(seq_lens) > 1:
            raise Exception(
                "Sequence must be same length but found lengths {}".format(
                    [len(s) for s in sequences]
                )
            )

        self.annotations = []

        self.window = window
        self._sequences = tuple([str(s) for s in sequences])
        if sequence_labels is None:
            sequence_labels = [""] * len(sequences)
        for i in apply_indices:
            sequence_labels[i] = "{index} " + sequence_labels[i]

        if foreground_colors == self.RANDOM_COLOR:
            foreground_colors = [random_color() for _ in sequences]
        self.foreground_colors = foreground_colors
        if background_colors == self.RANDOM_COLOR:
            background_colors = [random_color() for _ in sequences]
        self.background_colors = background_colors

        self.sequence_labels = sequence_labels
        self.indent = indent
        self.width = width
        self.spacer = spacer
        self.header_spacer = header_spacer
        if name is None:
            name = self.DEFAULTS.NAME
        self.name = name
        self.metadata = OrderedDict()
        if description:
            self.metadata["Description"] = self.DEFAULTS.DESCRIPTION
        if hasattr(self.sequences[0], "cyclic"):
            self.metadata["Cyclic"] = self.sequences[0].cyclic
        if metadata is not None:
            self.metadata.update(metadata)

    def set_window(self, start, end):
        """Sets the inclusive viewing window"""
        self.window = (start, end)
        return self

    @property
    def sequences(self):
        return list(self._sequences)

    @property
    def header(self):
        """Return the formatted header and metadata"""
        metadata = "\n".join(
            "{key}: {val}".format(key=key, val=val)
            for key, val in self.metadata.items()
        )
        metadata = indent(metadata, self.DEFAULTS.METADATA_INDENT)
        return '> "{name}" ({length}bp)\n{metadata}'.format(
            name=self.name, length=len(self), metadata=metadata
        )

    @property
    def rows(self):
        lines = []
        for seq in self.sequences:
            line = to_lines(str(seq)[self.window[0] : self.window[1]], width=self.width)
            lines.append(line)
        interleafed = functools.reduce(lambda x, y: x + y, zip(*lines))
        chunks = chunkify(interleafed, len(self.sequences))
        rows = []
        index = self.window[0]
        for chunk in chunks:
            labels = [str(l).format(index=index) for l in self.sequence_labels]
            rows.append(
                SequenceRow(
                    chunk,
                    labels,
                    self.indent,
                    index,
                    min(index + self.width - 1, len(self)),
                    line_colors=self.foreground_colors,
                    line_backgrounds=self.background_colors,
                )
            )
            index += len(chunk[0])
        self._annotate_rows(rows)
        return rows

    def annotate(
        self, start, end, label=None, fill=None, color=None, background=None, top=True
    ):
        """
        Annotates this viewer object starting from 'start' to 'end' inclusively.

        :param start: inclusive start
        :type start: int
        :param end: inclusive end
        :type end: int
        :param label: optional label to apply to the annotation
        :type label: basestring | StringColumn
        :param fill: the fill character to use to (e.g. '<', '>', '^') to fill in whitespace
        :type fill: string
        :param color: the foreground color to apply to the annotation (hex or common name)
        :type color: string
        :param background: the foreground color to apply to the annotation (hex or common name)
        :type background: string
        :return: None
        :rtype: None
        """
        if fill is None:
            fill = ViewerAnnotationFlag.BOTH
        if label is None:
            label = ""
        self.annotations.append(
            dict(
                start=start,
                end=end,
                label=label,
                fill=str(fill),
                color=color,
                background=background,
                top=top,
            )
        )

    def _annotate_rows(self, rows):
        """Annotate the rows using the viewer's annotations"""
        for a in self.annotations:
            for row in rows:
                if a["end"] >= row.start and a["start"] <= row.end:
                    row.annotate(**a)
        return rows

    def print(self):
        print(str(self))

    def __len__(self):
        return len(self.sequences[0])

    def __str__(self):
        spacer = self.spacer
        if spacer is None:
            spacer = ""
        s = "{header}\n".format(header=self.header)
        s += self.header_spacer
        s += "\n{}".format(spacer).join([str(r) for r in self.rows])
        return s


class FASTAItem(SequenceViewer):
    def __init__(self, sequence):
        super().__init__(
            [sequence],
            indent=0,
            width=80,
            name=sequence.name,
            apply_indices=[],
            sequence_labels=[""],
            spacer="",
            header_spacer="",
        )

    @property
    def header(self):
        return ">{}".format(self.name)


class FASTAViewer(object):
    def __init__(self, sequences):
        self.views = [FASTAItem(sequence) for sequence in sequences]

    def __str__(self):
        return "\n\n".join(str(v) for v in self.views)

    def print(self):
        print(str(self))



class ContigContainerError(Exception):
    """Exception for ContigContainer"""


class ContigError(Exception):
    """Exception for Contig"""


class RegionError(Exception):
    """Exception for Region"""


class SequenceError(Exception):
    """Exception for Sequence handling"""
